from gwatn.data_classes.d_edge import DEdge
from gwatn.data_classes.d_node import DNode
from gwatn.data_classes.market_type import MarketType


__all__ = [
    "DEdge",
    "DNode",
    "MarketType",
]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class DistributionTariff(StrEnum):
    """
    Name of distribution tariff of local network company/utility

    Choices and descriptions:

      * Unknown:
      * VersantStorageHeatTariff:
      * VersantATariff:
    """

    Unknown = auto()
    VersantStorageHeatTariff = auto()
    VersantATariff = auto()

    @classmethod
    def default(cls) -> "DistributionTariff":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class ActorClass(StrEnum):
    """
    Determines code running Spaceheat Node supervised by SCADA

    Choices and descriptions:

      * NoActor:
      * PrimaryScada:
      * PrimaryMeter:
      * BooleanActuator:
      * SimpleSensor:
      * HomeAlone:
      * Atn:
      * MultipurposeSensor:
      * Thermostat:
    """

    NoActor = auto()
    PrimaryScada = auto()
    PrimaryMeter = auto()
    BooleanActuator = auto()
    SimpleSensor = auto()
    HomeAlone = auto()
    Atn = auto()
    MultipurposeSensor = auto()
    Thermostat = auto()

    @classmethod
    def default(cls) -> "ActorClass":
        """
        Returns default value NoActor
        """
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from gwproto.messages import GtDispatchBoolean_Maker
from gwproto.messages import GtShStatus_Maker
from gwproto.messages import GtTelemetry_Maker
from gwproto.messages import SnapshotSpaceheat_Maker

from gwatn.api_types import TypeMakerByName


HackTypeMakerByName = dict(TypeMakerByName)

HackTypeMakerByName[GtDispatchBoolean_Maker.type_alias] = GtDispatchBoolean_Maker
HackTypeMakerByName[GtShStatus_Maker.type_alias] = GtShStatus_Maker
HackTypeMakerByName[SnapshotSpaceheat_Maker.type_alias] = SnapshotSpaceheat_Maker
HackTypeMakerByName[GtTelemetry_Maker.type_alias] = GtTelemetry_Maker



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class AtnSpaceheatStrategyName(StrEnum):
    """


    Choices and descriptions:

      * NoActor:
      * SupervisorA:
      * HeatPumpWithBoostStore:
    """

    NoActor = auto()
    SupervisorA = auto()
    HeatPumpWithBoostStore = auto()

    @classmethod
    def default(cls) -> "AtnSpaceheatStrategyName":
        """
        Returns default value NoActor
        """
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type join.dispatch.contract, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

import gridworks.algo_utils as algo_utils
import gridworks.api_utils as api_utils
import gridworks.gw_config as config
from algosdk import encoding
from algosdk.future import transaction
from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class JoinDispatchContract(BaseModel):
    """Sent from a Scada to its paired AtomicTNode on RabbitMQ.

    This is sent as an invitation to join the DispatchContract. Upon receipt of this,
    the AtomicTNode can check that the DispatchContract has finished the first part of
    its bootstrapping. This means it is well-funded, and also has the Scada Cert Id
    and the Scada Addr publicly available. The AtomicTNode can check these against
    the signature provided by the SCADA in its invitation. An AtomicTNode actor accepts
    the invitation by finishing the Dispatch Contract bootstrap (which it can only do
    if its Algorand Account holds the associated TaTradingRights certificate) and
    then responding to the SCADA via RabbitMQ with a dispatch.contract.confirmed payload.

    https://gridworks.readthedocs.io/en/latest/dispatch-contract.html
    """

    FromGNodeAlias: str = Field(
        title="FromGNodeAlias",
    )
    FromGNodeInstanceId: str = Field(
        title="FromGNodeInstanceId",
    )
    DispatchContractAppId: int = Field(
        title="DispatchContractAppId",
    )
    SignedProof: str = Field(
        title="SignedProof",
    )
    TypeName: Literal["join.dispatch.contract"] = "join.dispatch.contract"
    Version: str = "000"

    @validator("FromGNodeAlias")
    def _check_from_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("FromGNodeInstanceId")
    def _check_from_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("SignedProof")
    def _check_signed_proof(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedProof failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator
    def check_axiom_0(cls, v: dict) -> dict:
        """
        Axiom 0: ScadaCert matches FromGNodeAlias.
        The name in the ScadaCert should be the GNodeAlias of the TerminalAsset corresponding
        to the sending SCADA. Therefore, FromGNodeAlias should be equal to the name of the
        ScadaCert ASA with `.scada` appended.
        """
        txn = encoding.future_msgpack_decode(v.get("SignedProof", None))
        DispatchContractAppId = v.get("DispatchContractAppId", None)
        FromGNodeAlias = v.get("FromGNodeAlias", None)

        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class JoinDispatchContract_Maker:
    type_name = "join.dispatch.contract"
    version = "000"

    def __init__(
        self,
        from_g_node_alias: str,
        from_g_node_instance_id: str,
        dispatch_contract_app_id: int,
        signed_proof: str,
    ):
        self.tuple = JoinDispatchContract(
            FromGNodeAlias=from_g_node_alias,
            FromGNodeInstanceId=from_g_node_instance_id,
            DispatchContractAppId=dispatch_contract_app_id,
            SignedProof=signed_proof,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: JoinDispatchContract) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> JoinDispatchContract:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> JoinDispatchContract:
        d2 = dict(d)
        if "FromGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeAlias")
        if "FromGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeInstanceId")
        if "DispatchContractAppId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DispatchContractAppId")
        if "SignedProof" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedProof")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return JoinDispatchContract(
            FromGNodeAlias=d2["FromGNodeAlias"],
            FromGNodeInstanceId=d2["FromGNodeInstanceId"],
            DispatchContractAppId=d2["DispatchContractAppId"],
            SignedProof=d2["SignedProof"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class AlgoCertType(StrEnum):
    """
    Used to distinguish ASA vs SmartSignature certificates

    Choices and descriptions:

      * ASA: Certificate based on Algorand Standard Asset
      * SmartSig: Certificate based on Algorand Smart Signature
    """

    ASA = auto()
    SmartSig = auto()

    @classmethod
    def default(cls) -> "AlgoCertType":
        """
        Returns default value ASA
        """
        return cls.ASA

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class GniStatus(StrEnum):
    """
    Enum for managing GNodeInstance lifecycle. [More Info](https://gridworks.readthedocs.io/en/latest/g-node-instance.html).

    Choices and descriptions:

      * Unknown: Default Value
      * Pending: Has been created by the World, but has not yet finished provisioning
      * Active: Active in its GridWorks world. If the GNodeInstance has an actor, that actor is communicating
      * Done: No longer represents the GNode.
    """

    Unknown = auto()
    Pending = auto()
    Active = auto()
    Done = auto()

    @classmethod
    def default(cls) -> "GniStatus":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type discoverycert.algo.create, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.enums import CoreGNodeRole


class CoreGNodeRole000SchemaEnum:
    enum_name: str = "core.g.node.role.000"
    symbols: List[str] = [
        "00000000",
        "0f8872f7",
        "d9823442",
        "86f21dd2",
        "9521af06",
        "4502e355",
        "d67e564e",
        "7a8e4046",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class CoreGNodeRole000(StrEnum):
    Other = auto()
    TerminalAsset = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    Scada = auto()

    @classmethod
    def default(cls) -> "CoreGNodeRole000":
        return cls.Other

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class CoreGNodeRoleMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> CoreGNodeRole:
        if not CoreGNodeRole000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to CoreGNodeRole000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, CoreGNodeRole, CoreGNodeRole.default())

    @classmethod
    def local_to_type(cls, core_g_node_role: CoreGNodeRole) -> str:
        if not isinstance(core_g_node_role, CoreGNodeRole):
            raise SchemaError(f"{core_g_node_role} must be of type {CoreGNodeRole}")
        versioned_enum = as_enum(
            core_g_node_role, CoreGNodeRole000, CoreGNodeRole000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, CoreGNodeRole000] = {
        "00000000": CoreGNodeRole000.Other,
        "0f8872f7": CoreGNodeRole000.TerminalAsset,
        "d9823442": CoreGNodeRole000.AtomicTNode,
        "86f21dd2": CoreGNodeRole000.MarketMaker,
        "9521af06": CoreGNodeRole000.AtomicMeteringNode,
        "4502e355": CoreGNodeRole000.ConductorTopologyNode,
        "d67e564e": CoreGNodeRole000.InterconnectionComponent,
        "7a8e4046": CoreGNodeRole000.Scada,
    }

    versioned_enum_to_type_dict: Dict[CoreGNodeRole000, str] = {
        CoreGNodeRole000.Other: "00000000",
        CoreGNodeRole000.TerminalAsset: "0f8872f7",
        CoreGNodeRole000.AtomicTNode: "d9823442",
        CoreGNodeRole000.MarketMaker: "86f21dd2",
        CoreGNodeRole000.AtomicMeteringNode: "9521af06",
        CoreGNodeRole000.ConductorTopologyNode: "4502e355",
        CoreGNodeRole000.InterconnectionComponent: "d67e564e",
        CoreGNodeRole000.Scada: "7a8e4046",
    }


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


class DiscoverycertAlgoCreate(BaseModel):
    """ """

    GNodeAlias: str = Field(
        title="GNodeAlias",
    )
    Role: CoreGNodeRole = Field(
        title="Role",
    )
    OldChildAliasList: List[str] = Field(
        title="OldChildAliasList",
    )
    DiscovererAddr: str = Field(
        title="DiscovererAddr",
    )
    SupportingMaterialHash: str = Field(
        title="SupportingMaterialHash",
    )
    MicroLat: Optional[int] = Field(
        title="MicroLat",
        default=None,
    )
    MicroLon: Optional[int] = Field(
        title="MicroLon",
        default=None,
    )
    TypeName: Literal["discoverycert.algo.create"] = "discoverycert.algo.create"
    Version: str = "000"

    @validator("GNodeAlias")
    def _check_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"GNodeAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("Role")
    def _check_role(cls, v: CoreGNodeRole) -> CoreGNodeRole:
        return as_enum(v, CoreGNodeRole, CoreGNodeRole.Other)

    @validator("OldChildAliasList")
    def _check_old_child_alias_list(cls, v: List) -> List:
        for elt in v:
            try:
                check_is_left_right_dot(elt)
            except ValueError as e:
                raise ValueError(
                    f"OldChildAliasList element {elt} failed LeftRightDot format validation: {e}"
                )
        return v

    @validator("DiscovererAddr")
    def _check_discoverer_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"DiscovererAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Role"]
        Role = as_enum(self.Role, CoreGNodeRole, CoreGNodeRole.default())
        d["RoleGtEnumSymbol"] = CoreGNodeRoleMap.local_to_type(Role)
        if d["MicroLat"] is None:
            del d["MicroLat"]
        if d["MicroLon"] is None:
            del d["MicroLon"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class DiscoverycertAlgoCreate_Maker:
    type_name = "discoverycert.algo.create"
    version = "000"

    def __init__(
        self,
        g_node_alias: str,
        role: CoreGNodeRole,
        old_child_alias_list: List[str],
        discoverer_addr: str,
        supporting_material_hash: str,
        micro_lat: Optional[int],
        micro_lon: Optional[int],
    ):
        self.tuple = DiscoverycertAlgoCreate(
            GNodeAlias=g_node_alias,
            Role=role,
            OldChildAliasList=old_child_alias_list,
            DiscovererAddr=discoverer_addr,
            SupportingMaterialHash=supporting_material_hash,
            MicroLat=micro_lat,
            MicroLon=micro_lon,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: DiscoverycertAlgoCreate) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> DiscoverycertAlgoCreate:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> DiscoverycertAlgoCreate:
        d2 = dict(d)
        if "GNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeAlias")
        if "RoleGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RoleGtEnumSymbol")
        if d2["RoleGtEnumSymbol"] in CoreGNodeRole000SchemaEnum.symbols:
            d2["Role"] = CoreGNodeRoleMap.type_to_local(d2["RoleGtEnumSymbol"])
        else:
            d2["Role"] = CoreGNodeRole.default()
        if "OldChildAliasList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing OldChildAliasList")
        if "DiscovererAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DiscovererAddr")
        if "SupportingMaterialHash" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupportingMaterialHash")
        if "MicroLat" not in d2.keys():
            d2["MicroLat"] = None
        if "MicroLon" not in d2.keys():
            d2["MicroLon"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return DiscoverycertAlgoCreate(
            GNodeAlias=d2["GNodeAlias"],
            Role=d2["Role"],
            OldChildAliasList=d2["OldChildAliasList"],
            DiscovererAddr=d2["DiscovererAddr"],
            SupportingMaterialHash=d2["SupportingMaterialHash"],
            MicroLat=d2["MicroLat"],
            MicroLon=d2["MicroLon"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type initial.tadeed.algo.create, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class InitialTadeedAlgoCreate(BaseModel):
    """TaValidator sends to GNodeFactory to complete creation of an initial TaDeed.

    If this message is valid, the GNodeFactory co-signs and submits the TaDeed creation. In addition, the
    GnodeFactory creates a TerminalAsset with GNodeStatus pending.  For more information:
    [TaDeed](https://gridworks.readthedocs.io/en/latest/ta-deed.html)
    [TaValidator](https://gridworks.readthedocs.io/en/latest/ta-validator.html)
    """

    ValidatorAddr: str = Field(
        title="Address of the TaValidator",
        description="The Algorand address of the TaValidator who is going to validate the location, device type, and power metering of the TerminalAsset.",
    )
    HalfSignedDeedCreationMtx: str = Field(
        title="Algo mulit-transaction for TaDeed creation",
    )
    TypeName: Literal["initial.tadeed.algo.create"] = "initial.tadeed.algo.create"
    Version: str = "000"

    @validator("ValidatorAddr")
    def _check_validator_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("HalfSignedDeedCreationMtx")
    def _check_half_signed_deed_creation_mtx(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"HalfSignedDeedCreationMtx failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded HalfSignedDeedCreationMtx must have type MultisigTransaction from the
        2-sig MultiAccount  [GnfAdminAddr, ValidatorAddr].
        [More info](https://gridworks.readthedocs.io/en/latest/g-node-factory.html#gnfadminaddr)
        """
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: Creates Initial ASA TaDeed.
        The transaction must create an Algorand Standard Asset
           - Total is 1
           - UnitName is TADEED
           - Manager is GnfAdminAddr
           - AssetName has the following characteristics:
                 - length <=  32 characters
                 - LeftRightDot format
                 - final word is '.ta'
        [More info](https://gridworks.readthedocs.io/en/latest/ta-deed.html#asa-tadeed-specs)
        """
        return v

    @root_validator
    def check_axiom_3(cls, v: dict) -> dict:
        """
        Axiom 3: Mtx signed by TaValidator.

        """
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class InitialTadeedAlgoCreate_Maker:
    type_name = "initial.tadeed.algo.create"
    version = "000"

    def __init__(self, validator_addr: str, half_signed_deed_creation_mtx: str):
        self.tuple = InitialTadeedAlgoCreate(
            ValidatorAddr=validator_addr,
            HalfSignedDeedCreationMtx=half_signed_deed_creation_mtx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: InitialTadeedAlgoCreate) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> InitialTadeedAlgoCreate:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> InitialTadeedAlgoCreate:
        d2 = dict(d)
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "HalfSignedDeedCreationMtx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HalfSignedDeedCreationMtx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return InitialTadeedAlgoCreate(
            ValidatorAddr=d2["ValidatorAddr"],
            HalfSignedDeedCreationMtx=d2["HalfSignedDeedCreationMtx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type base.g.node.gt, version 002"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.data_classes.base_g_node import BaseGNode
from gwatn.enums import CoreGNodeRole
from gwatn.enums import GNodeStatus


class CoreGNodeRole000SchemaEnum:
    enum_name: str = "core.g.node.role.000"
    symbols: List[str] = [
        "00000000",
        "0f8872f7",
        "d9823442",
        "86f21dd2",
        "9521af06",
        "4502e355",
        "d67e564e",
        "7a8e4046",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class CoreGNodeRole000(StrEnum):
    Other = auto()
    TerminalAsset = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    Scada = auto()

    @classmethod
    def default(cls) -> "CoreGNodeRole000":
        return cls.Other

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class CoreGNodeRoleMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> CoreGNodeRole:
        if not CoreGNodeRole000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to CoreGNodeRole000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, CoreGNodeRole, CoreGNodeRole.default())

    @classmethod
    def local_to_type(cls, core_g_node_role: CoreGNodeRole) -> str:
        if not isinstance(core_g_node_role, CoreGNodeRole):
            raise SchemaError(f"{core_g_node_role} must be of type {CoreGNodeRole}")
        versioned_enum = as_enum(
            core_g_node_role, CoreGNodeRole000, CoreGNodeRole000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, CoreGNodeRole000] = {
        "00000000": CoreGNodeRole000.Other,
        "0f8872f7": CoreGNodeRole000.TerminalAsset,
        "d9823442": CoreGNodeRole000.AtomicTNode,
        "86f21dd2": CoreGNodeRole000.MarketMaker,
        "9521af06": CoreGNodeRole000.AtomicMeteringNode,
        "4502e355": CoreGNodeRole000.ConductorTopologyNode,
        "d67e564e": CoreGNodeRole000.InterconnectionComponent,
        "7a8e4046": CoreGNodeRole000.Scada,
    }

    versioned_enum_to_type_dict: Dict[CoreGNodeRole000, str] = {
        CoreGNodeRole000.Other: "00000000",
        CoreGNodeRole000.TerminalAsset: "0f8872f7",
        CoreGNodeRole000.AtomicTNode: "d9823442",
        CoreGNodeRole000.MarketMaker: "86f21dd2",
        CoreGNodeRole000.AtomicMeteringNode: "9521af06",
        CoreGNodeRole000.ConductorTopologyNode: "4502e355",
        CoreGNodeRole000.InterconnectionComponent: "d67e564e",
        CoreGNodeRole000.Scada: "7a8e4046",
    }


class GNodeStatus100SchemaEnum:
    enum_name: str = "g.node.status.100"
    symbols: List[str] = [
        "00000000",
        "153d3475",
        "a2cfc2f7",
        "839b38db",
        "f5831e1d",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class GNodeStatus100(StrEnum):
    Unknown = auto()
    Pending = auto()
    Active = auto()
    PermanentlyDeactivated = auto()
    Suspended = auto()

    @classmethod
    def default(cls) -> "GNodeStatus100":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class GNodeStatusMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> GNodeStatus:
        if not GNodeStatus100SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to GNodeStatus100 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, GNodeStatus, GNodeStatus.default())

    @classmethod
    def local_to_type(cls, g_node_status: GNodeStatus) -> str:
        if not isinstance(g_node_status, GNodeStatus):
            raise SchemaError(f"{g_node_status} must be of type {GNodeStatus}")
        versioned_enum = as_enum(
            g_node_status, GNodeStatus100, GNodeStatus100.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, GNodeStatus100] = {
        "00000000": GNodeStatus100.Unknown,
        "153d3475": GNodeStatus100.Pending,
        "a2cfc2f7": GNodeStatus100.Active,
        "839b38db": GNodeStatus100.PermanentlyDeactivated,
        "f5831e1d": GNodeStatus100.Suspended,
    }

    versioned_enum_to_type_dict: Dict[GNodeStatus100, str] = {
        GNodeStatus100.Unknown: "00000000",
        GNodeStatus100.Pending: "153d3475",
        GNodeStatus100.Active: "a2cfc2f7",
        GNodeStatus100.PermanentlyDeactivated: "839b38db",
        GNodeStatus100.Suspended: "f5831e1d",
    }


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


class BaseGNodeGt(BaseModel):
    """.

    BaseGNode. Authority is GNodeFactory.
    """

    GNodeId: str = Field(
        title="GNodeId",
    )
    Alias: str = Field(
        title="Alias",
    )
    Status: GNodeStatus = Field(
        title="Status",
    )
    Role: CoreGNodeRole = Field(
        title="Role",
    )
    GNodeRegistryAddr: str = Field(
        title="GNodeRegistryAddr",
    )
    PrevAlias: Optional[str] = Field(
        title="PrevAlias",
        default=None,
    )
    GpsPointId: Optional[str] = Field(
        title="GpsPointId",
        default=None,
    )
    OwnershipDeedId: Optional[int] = Field(
        title="OwnershipDeedId",
        default=None,
    )
    OwnershipDeedValidatorAddr: Optional[str] = Field(
        title="OwnershipDeedValidatorAddr",
        default=None,
    )
    OwnerAddr: Optional[str] = Field(
        title="OwnerAddr",
        default=None,
    )
    DaemonAddr: Optional[str] = Field(
        title="DaemonAddr",
        default=None,
    )
    TradingRightsId: Optional[int] = Field(
        title="TradingRightsId",
        default=None,
    )
    ScadaAlgoAddr: Optional[str] = Field(
        title="ScadaAlgoAddr",
        default=None,
    )
    ScadaCertId: Optional[int] = Field(
        title="ScadaCertId",
        default=None,
    )
    TypeName: Literal["base.g.node.gt"] = "base.g.node.gt"
    Version: str = "002"

    @validator("GNodeId")
    def _check_g_node_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("Alias")
    def _check_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"Alias failed LeftRightDot format validation: {e}")
        return v

    @validator("Status")
    def _check_status(cls, v: GNodeStatus) -> GNodeStatus:
        return as_enum(v, GNodeStatus, GNodeStatus.Unknown)

    @validator("Role")
    def _check_role(cls, v: CoreGNodeRole) -> CoreGNodeRole:
        return as_enum(v, CoreGNodeRole, CoreGNodeRole.Other)

    @validator("GNodeRegistryAddr")
    def _check_g_node_registry_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeRegistryAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("PrevAlias")
    def _check_prev_alias(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"PrevAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("GpsPointId")
    def _check_gps_point_id(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GpsPointId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("OwnershipDeedValidatorAddr")
    def _check_ownership_deed_validator_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"OwnershipDeedValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("OwnerAddr")
    def _check_owner_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"OwnerAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("DaemonAddr")
    def _check_daemon_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"DaemonAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("ScadaAlgoAddr")
    def _check_scada_algo_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ScadaAlgoAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Status"]
        Status = as_enum(self.Status, GNodeStatus, GNodeStatus.default())
        d["StatusGtEnumSymbol"] = GNodeStatusMap.local_to_type(Status)
        del d["Role"]
        Role = as_enum(self.Role, CoreGNodeRole, CoreGNodeRole.default())
        d["RoleGtEnumSymbol"] = CoreGNodeRoleMap.local_to_type(Role)
        if d["PrevAlias"] is None:
            del d["PrevAlias"]
        if d["GpsPointId"] is None:
            del d["GpsPointId"]
        if d["OwnershipDeedId"] is None:
            del d["OwnershipDeedId"]
        if d["OwnershipDeedValidatorAddr"] is None:
            del d["OwnershipDeedValidatorAddr"]
        if d["OwnerAddr"] is None:
            del d["OwnerAddr"]
        if d["DaemonAddr"] is None:
            del d["DaemonAddr"]
        if d["TradingRightsId"] is None:
            del d["TradingRightsId"]
        if d["ScadaAlgoAddr"] is None:
            del d["ScadaAlgoAddr"]
        if d["ScadaCertId"] is None:
            del d["ScadaCertId"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class BaseGNodeGt_Maker:
    type_name = "base.g.node.gt"
    version = "002"

    def __init__(
        self,
        g_node_id: str,
        alias: str,
        status: GNodeStatus,
        role: CoreGNodeRole,
        g_node_registry_addr: str,
        prev_alias: Optional[str],
        gps_point_id: Optional[str],
        ownership_deed_id: Optional[int],
        ownership_deed_validator_addr: Optional[str],
        owner_addr: Optional[str],
        daemon_addr: Optional[str],
        trading_rights_id: Optional[int],
        scada_algo_addr: Optional[str],
        scada_cert_id: Optional[int],
    ):
        self.tuple = BaseGNodeGt(
            GNodeId=g_node_id,
            Alias=alias,
            Status=status,
            Role=role,
            GNodeRegistryAddr=g_node_registry_addr,
            PrevAlias=prev_alias,
            GpsPointId=gps_point_id,
            OwnershipDeedId=ownership_deed_id,
            OwnershipDeedValidatorAddr=ownership_deed_validator_addr,
            OwnerAddr=owner_addr,
            DaemonAddr=daemon_addr,
            TradingRightsId=trading_rights_id,
            ScadaAlgoAddr=scada_algo_addr,
            ScadaCertId=scada_cert_id,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: BaseGNodeGt) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> BaseGNodeGt:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> BaseGNodeGt:
        d2 = dict(d)
        if "GNodeId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeId")
        if "Alias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Alias")
        if "StatusGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StatusGtEnumSymbol")
        if d2["StatusGtEnumSymbol"] in GNodeStatus100SchemaEnum.symbols:
            d2["Status"] = GNodeStatusMap.type_to_local(d2["StatusGtEnumSymbol"])
        else:
            d2["Status"] = GNodeStatus.default()
        if "RoleGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RoleGtEnumSymbol")
        if d2["RoleGtEnumSymbol"] in CoreGNodeRole000SchemaEnum.symbols:
            d2["Role"] = CoreGNodeRoleMap.type_to_local(d2["RoleGtEnumSymbol"])
        else:
            d2["Role"] = CoreGNodeRole.default()
        if "GNodeRegistryAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeRegistryAddr")
        if "PrevAlias" not in d2.keys():
            d2["PrevAlias"] = None
        if "GpsPointId" not in d2.keys():
            d2["GpsPointId"] = None
        if "OwnershipDeedId" not in d2.keys():
            d2["OwnershipDeedId"] = None
        if "OwnershipDeedValidatorAddr" not in d2.keys():
            d2["OwnershipDeedValidatorAddr"] = None
        if "OwnerAddr" not in d2.keys():
            d2["OwnerAddr"] = None
        if "DaemonAddr" not in d2.keys():
            d2["DaemonAddr"] = None
        if "TradingRightsId" not in d2.keys():
            d2["TradingRightsId"] = None
        if "ScadaAlgoAddr" not in d2.keys():
            d2["ScadaAlgoAddr"] = None
        if "ScadaCertId" not in d2.keys():
            d2["ScadaCertId"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return BaseGNodeGt(
            GNodeId=d2["GNodeId"],
            Alias=d2["Alias"],
            Status=d2["Status"],
            Role=d2["Role"],
            GNodeRegistryAddr=d2["GNodeRegistryAddr"],
            PrevAlias=d2["PrevAlias"],
            GpsPointId=d2["GpsPointId"],
            OwnershipDeedId=d2["OwnershipDeedId"],
            OwnershipDeedValidatorAddr=d2["OwnershipDeedValidatorAddr"],
            OwnerAddr=d2["OwnerAddr"],
            DaemonAddr=d2["DaemonAddr"],
            TradingRightsId=d2["TradingRightsId"],
            ScadaAlgoAddr=d2["ScadaAlgoAddr"],
            ScadaCertId=d2["ScadaCertId"],
            TypeName=d2["TypeName"],
            Version="002",
        )

    @classmethod
    def tuple_to_dc(cls, t: BaseGNodeGt) -> BaseGNode:
        if t.GNodeId in BaseGNode.by_id.keys():
            dc = BaseGNode.by_id[t.GNodeId]
        else:
            dc = BaseGNode(
                g_node_id=t.GNodeId,
                alias=t.Alias,
                status=t.Status,
                role=t.Role,
                g_node_registry_addr=t.GNodeRegistryAddr,
                prev_alias=t.PrevAlias,
                gps_point_id=t.GpsPointId,
                ownership_deed_id=t.OwnershipDeedId,
                ownership_deed_validator_addr=t.OwnershipDeedValidatorAddr,
                owner_addr=t.OwnerAddr,
                daemon_addr=t.DaemonAddr,
                trading_rights_id=t.TradingRightsId,
                scada_algo_addr=t.ScadaAlgoAddr,
                scada_cert_id=t.ScadaCertId,
            )

        return dc

    @classmethod
    def dc_to_tuple(cls, dc: BaseGNode) -> BaseGNodeGt:
        t = BaseGNodeGt_Maker(
            g_node_id=dc.g_node_id,
            alias=dc.alias,
            status=dc.status,
            role=dc.role,
            g_node_registry_addr=dc.g_node_registry_addr,
            prev_alias=dc.prev_alias,
            gps_point_id=dc.gps_point_id,
            ownership_deed_id=dc.ownership_deed_id,
            ownership_deed_validator_addr=dc.ownership_deed_validator_addr,
            owner_addr=dc.owner_addr,
            daemon_addr=dc.daemon_addr,
            trading_rights_id=dc.trading_rights_id,
            scada_algo_addr=dc.scada_algo_addr,
            scada_cert_id=dc.scada_cert_id,
        ).tuple
        return t

    @classmethod
    def type_to_dc(cls, t: str) -> BaseGNode:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: BaseGNode) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> BaseGNode:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))



"""Settings for an AtomicTNode, readable from environment and/or from env files."""

from typing import Optional

import pendulum
from gridworks.gw_config import AlgoApiSecrets
from gridworks.gw_config import GNodeSettings
from gridworks.gw_config import Public
from pydantic import BaseSettings
from pydantic import SecretStr


class AtnSettings(GNodeSettings):
    g_node_alias: str = "d1.isone.ver.keene.holly"
    g_node_role_value: str = "AtomicTNode"
    my_super_alias: str = "d1.isone.ver.keene.super1"
    g_node_id: str = "6bb37cc5-740d-40f5-a535-43987a5d07b4"
    g_node_instance_id: str = "00000000-0000-0000-0000-000000000000"
    scada_gni_id: Optional[str] = None
    sk: SecretStr = SecretStr(
        "K6iB3AHmzSQ8wDE91QdUfaheDMEtf2WJUMYeeRptKxHiTxG3HC+iKpngXmi82y2r9uVPYwTI5aGiMhdXmPRxcQ=="
    )
    # Public address 4JHRDNY4F6RCVGPALZULZWZNVP3OKT3DATEOLINCGILVPGHUOFY7KCHVIQ

    # Next 4 settings are consistent with dev env settings in gridworks-marketmaker repo
    market_maker_alias = "d1.isone.ver.keene"
    market_maker_algo_address = (
        "CYWMWYHJ7ON4IR5XQDJBBPDU472QU4KJQ6XQVZBIIRTCHT6SHTFNHEAVC4"
    )
    mm_api_root = "http://localhost:7997"
    initial_time_unix_s = pendulum.datetime(
        year=2020, month=1, day=1, hour=4, minute=20
    ).int_timestamp

    class Config:
        env_prefix = "ATN_"
        env_nested_delimiter = "__"


class SupervisorSettings(GNodeSettings):
    g_node_alias: str = "d1.isone.ver.keene.super1"
    g_node_role_value: str = "Supervisor"
    supervisor_container_id: str = "3d9ad9fa-8e0c-4569-8f34-2b0198a55085"

    class Config:
        env_prefix = "SUPER_"
        env_nested_delimiter = "__"


class ValidatorSettings(BaseSettings):
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    public: Public = Public()
    sk: SecretStr = SecretStr(
        "FCLmrvflibLD6Deu3NNiUQCC9LOWpXLsbMR/cP2oJzH8IT4Zu8vBAhRNsXoWF+2i6q2KyBZrPhmbDCKJD7rBBg=="
    )
    cert_name: str = "Molly Metermaid"
    name: str = "Molly Inc Telemetry Surveyors and Purveyors"
    api_root: str = "http://localhost:8001"

    class Config:
        env_prefix = "VLDTR_"
        env_nested_delimiter = "__"


class TaOwnerSettings(BaseSettings):
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    public: Public = Public()
    sk: SecretStr = SecretStr(
        "sp4SDWmH8Rin0IhPJQq1UMsSR5C0j1IGqzLdcwCMySBVzT8lEUwjwwpS9z6l6dKSg52WWEjRdJDAL+eVt4kvBg=="
    )

    ta_daemon_addr: str = "NZXUSTZACPVJBHRSSJ5KE3JUPCITK5P2O4FE67NYPXRDVCJA6ZX4AL62EA"
    ta_validator_addr: str = (
        "7QQT4GN3ZPAQEFCNWF5BMF7NULVK3CWICZVT4GM3BQRISD52YEDLWJ4MII"
    )
    ta_validator_api_root: str = "http://localhost:8001"
    ta_daemon_api_fqdn: str = "http://localhost"
    ta_daemon_api_port: str = "8002"
    initial_ta_alias: str = "d1.isone.ver.keene.holly.ta"
    micro_lat: int = 45511230
    micro_lon: int = -68354650

    class Config:
        env_prefix = "HH_"
        env_nested_delimiter = "__"


class TaDaemonSettings(BaseSettings):
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    public: Public = Public()
    sk: SecretStr = SecretStr(
        "tQ8ABbLLR96cnRE3Y2tlrj2d/rNPRFkf8FosJ46tVIlub0lPIBPqkJ4yknqibTR4kTV1+ncKT324feI6iSD2bw=="
    )
    ta_owner_addr: str = "KXGT6JIRJQR4GCSS647KL2OSSKBZ3FSYJDIXJEGAF7TZLN4JF4DGDDX4BI"
    validator_addr: str = "7QQT4GN3ZPAQEFCNWF5BMF7NULVK3CWICZVT4GM3BQRISD52YEDLWJ4MII"

    class Config:
        env_prefix = "TAD_"
        env_nested_delimiter = "__"


class ScadaSettings(GNodeSettings):
    cert_type_value: str = "ASA"
    g_node_alias: str = "d1.isone.ver.keene.holly.scada"
    g_node_role_value: str = "Scada"
    my_super_alias: str = "d1.super1"
    g_node_id: str = "282b2121-ec42-44e0-99fe-d77c40f64ac0"
    g_node_instance_id: str = "b6eab35e-ff90-4380-a5c9-08cf26aaf45e"
    atn_gni_id: Optional[str] = None
    # Next 4 settings are consistent with dev env settings in gridworks-marketmaker repo
    sk: SecretStr = SecretStr(
        "7RaOTVzqiL5NHarfkp7FpR1rjmn3T4Nibd+017Uhd9GpOVe2and91D46zJxO1wycU02TXxVu18cmslBEKgb8IA=="
    )
    # addr "MWKXTOQU3MQXCMJEIYD5PJGIR2RJNEUDVFIHLY75XZ5PJSY2H73W4DPHTM"
    initial_time_unix_s = pendulum.datetime(
        year=2020, month=1, day=1, hour=4, minute=20
    ).int_timestamp

    class Config:
        env_prefix = "SCADA_"
        env_nested_delimiter = "__"


class DiscovererSettings(BaseSettings):
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    public: Public = Public()
    sk: SecretStr = SecretStr(
        "X20eXB/VZilEmzaPCDSn9WsuGZ5/f0+IxuEhfYfVtmZR9q5bcbjpBodPpiUCCkr0Xv11sKYxf08PnAKQFNtW3Q=="
    )
    discovered_ctn_alias = "d1.isone.ver.keene.pwrs"
    original_child_alias_list = ["d1.isone.ver.keene.holly"]
    micro_lat = 44838681
    micro_lon = -68705311

    class Config:
        env_prefix = "DISC_"
        env_nested_delimiter = "__"



""" SCADA Actor """
import functools
import logging
import random
import time
from typing import Optional
from typing import no_type_check

import dotenv
import gridworks.algo_utils as algo_utils
import gridworks.conversion_factors as cf
import pendulum
import requests
from algosdk.atomic_transaction_composer import TransactionWithSigner
from algosdk.future.transaction import *
from algosdk.v2client.algod import AlgodClient
from beaker.client import ApplicationClient
from gridworks.actor_base import ActorBase
from gridworks.algo_utils import BasicAccount
from gridworks.enums import GNodeRole
from gridworks.message import as_enum
from pydantic import BaseModel

import gwatn.api_types as api_types
import gwatn.config as config
from gwatn import DispatchContract
from gwatn.enums import AlgoCertType
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.types import AtnParamsHeatpumpwithbooststore as AtnParams
from gwatn.types import (
    DispatchContractConfirmedHeatpumpwithbooststore as DispatchContractConfirmed,
)
from gwatn.types import (
    DispatchContractConfirmedHeatpumpwithbooststore_Maker as DispatchContractConfirmed_Maker,
)
from gwatn.types import GtDispatchBoolean
from gwatn.types import GtDispatchBoolean_Maker
from gwatn.types import GwCertId
from gwatn.types import GwCertId_Maker
from gwatn.types import HeartbeatA
from gwatn.types import HeartbeatA_Maker
from gwatn.types import HeartbeatB
from gwatn.types import HeartbeatB_Maker
from gwatn.types import JoinDispatchContract_Maker
from gwatn.types import ScadaCertTransfer_Maker
from gwatn.types import SimScadaDriverReport
from gwatn.types import SimScadaDriverReport_Maker
from gwatn.types import SimTimestep
from gwatn.types import SimTimestep_Maker
from gwatn.types import SnapshotHeatpumpwithbooststore


DISPATCH_CONTRACT_REPORTING_ALGOS = 5

LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.INFO)


def get_max_store_kwh_th(params: AtnParams) -> float:
    return (
        cf.KWH_TH_PER_GALLON_PER_DEG_F
        * params.StoreSizeGallons
        * (params.MaxStoreTempF - params.ZeroPotentialEnergyWaterTempF)
    )


class AtnHbStatus(BaseModel):
    LastHeartbeatReceivedMs: int
    AtnLastHex: Optional[str] = None
    ScadaLastHex: str = "0"


class ScadaActor(ActorBase):
    def __init__(
        self,
        settings: config.ScadaSettings = config.ScadaSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        super().__init__(settings=settings)
        self.api_type_maker_by_name = (
            api_types.TypeMakerByName
        )  # overwrites base class to include types used in this repo
        self.settings = settings
        self.atn_gni_id = settings.atn_gni_id
        self.acct: BasicAccount = BasicAccount(settings.sk.get_secret_value())
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )

        self.sp = self.client.suggested_params()
        self.cert_id: Optional[GwCertId] = None
        self.talking_with: bool = False
        self.atn_hb_status = AtnHbStatus(
            LastHeartbeatReceivedMs=int(1000 * time.time())
        )
        # The dc_client (DispatchContract client) builds on top of the Algorand beaker ApplicationClient
        self.dc_client: Optional[ApplicationClient] = None
        self.dc_app_id: Optional[int] = None
        self.algo_init_check()
        self.universe_type = as_enum(
            self.settings.universe_type_value, UniverseType, UniverseType.default()
        )
        self._time: float = self.get_initial_time_s()

        self.boost_power_kw: float = 0
        self.heatpump_power_kw: float = 0
        self.cop: float = 0
        self.store_kwh: int = 0
        self.max_store_kwh: int = 0
        self.atn_params: Optional[AtnParams] = None

    @property
    def atn_alias(self):
        """Removes `scada` from the end of the SCADA's GNodeAlias"""
        return self.alias[:-6]

    @property
    def ta_alias(self):
        """Removes `scada` from the end of the SCADA's GNodeAlias"""
        return self.atn_alias + ".ta"

    def local_rabbit_startup(self) -> None:
        rjb = MessageCategorySymbol.rjb.value
        tc_alias_lrh = self.settings.time_coordinator_alias.replace(".", "-")
        binding = f"{rjb}.{tc_alias_lrh}.timecoordinator.sim-timestep"

        cb = functools.partial(self.on_timecoordinator_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "timecoordinatormic_tx", routing_key=binding, callback=cb
        )
        LOGGER.info(
            f"Queue {self.queue_name} bound to timecoordinatormic_tx with {binding} "
        )

    @no_type_check
    def on_timecoordinator_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    def time(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self._time
        else:
            return time.time()

    def get_initial_time_s(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self.settings.initial_time_unix_s
        else:
            return time.time()

    def prepare_for_death(self) -> None:
        self.actor_main_stopped = True

    ########################
    ## Receives
    ########################

    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatB
    ) -> None:
        if payload.TypeName == DispatchContractConfirmed_Maker.type_name:
            if from_role != GNodeRole.AtomicTNode:
                LOGGER.info(
                    f"Ignoring DispatchContractConfrimed from GNode with role {from_role}; expects AtomicTNode"
                )
            try:
                self.dispatch_contract_confirmed_received(payload)
            except:
                LOGGER.exception("Error in dispatch_contract_confirmed_received")
        elif payload.TypeName == GtDispatchBoolean_Maker.type_name:
            if from_role != GNodeRole.AtomicTNode:
                LOGGER.info(
                    f"Ignoring GtDispatchBooleanfrom GNode with role {from_role}; expects AtomicTNode"
                )
            try:
                self.dispatch_received(payload)
            except:
                LOGGER.exception("Error in dispatch_received")

        elif payload.TypeName == HeartbeatA_Maker.type_name:
            if from_role != GNodeRole.Supervisor:
                LOGGER.info(
                    f"Ignoring HeartbeatA from GNode with role {from_role}; expects Supervisor"
                )
            try:
                self.heartbeat_from_super(from_alias=from_alias, ping=payload)
            except:
                LOGGER.exception("Error in heartbeat_from_super")
        elif payload.TypeName == HeartbeatB_Maker.type_name:
            if from_role != GNodeRole.AtomicTNode:
                LOGGER.info(
                    f"Ignoring HeartbeatB from GNode with role {from_role}; expects AtomicTNode"
                )
            try:
                self.heartbeat_from_atn(ping=payload)
            except:
                LOGGER.exception("Error in heartbeat_from_atn")
        elif payload.TypeName == SimScadaDriverReport_Maker.type_name:
            try:
                self.sim_scada_driver_report_received(payload)
            except:
                LOGGER.exception("Error in sim_scada_driver_report_received")
        elif payload.TypeName == SimTimestep_Maker.type_name:
            if from_role != GNodeRole.TimeCoordinator:
                LOGGER.info(
                    f"Ignoring SimTimestep from GNode with role {from_role}; expects TimeCoordinator"
                )
            try:
                self.timestep_from_timecoordinator(payload)
            except:
                LOGGER.exception("Error in timestep_from_timecoordinator")

    def heartbeat_from_atn(self, ping: HeartbeatB) -> None:
        """
        This is the Scada's half of the DispatchContract Heartbeat pattern.
        It:
          - Checks that it has a DispatchContract (owns that SmartContract)
          - Checks the GNodeAlias and GNodeInstanceId to validate partner
          - Sends a reply HeartbeatB immediately back using a RabbitJsonDirect message
          - Sends an audit report of its action to the DispatchContract
            # TODO: save audit report for sending in a batch if SmartContract
            # exists but is not reachable (i.e. blockchain down)
        [more info](https://gridworks.readthedocs.io/en/latest/dispatch-contract.html)

        Args:
            payload (HeartbeatB): The latest heartbeat received from its
            AtomicTNode partner

        """
        self.ping = ping

        received_ms = int(time.time() * 1000)
        if not self.in_dispatch_contract():
            LOGGER.info(f"Not in Dispatch Contract. Ignoring")
        if ping.FromGNodeInstanceId != self.atn_gni_id:
            raise Exception("In dispatch contract but mismatched Atn Gni Id!")
        if self.atn_hb_status.AtnLastHex is not None:
            if ping.YourLastHex != self.atn_hb_status.ScadaLastHex:
                LOGGER.info("Received incorrect ping. Ignoring")
                return
        self.talking_with = True
        self.atn_hb_status.LastHeartbeatReceivedMs = int(time.time() * 1000)
        self.atn_hb_status.AtnLastHex = ping.MyHex
        self.atn_hb_status.ScadaLastHex = str(random.choice("0123456789abcdef"))
        LOGGER.info(f"Got {ping.MyHex}. Sending {self.atn_hb_status.ScadaLastHex}")
        pong = HeartbeatB_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            my_hex=self.atn_hb_status.ScadaLastHex,
            your_last_hex=self.atn_hb_status.AtnLastHex,
            last_received_time_unix_ms=self.atn_hb_status.LastHeartbeatReceivedMs,
            send_time_unix_ms=int(1000 * time.time()),
        ).tuple
        self.send_message(
            payload=pong, to_role=GNodeRole.AtomicTNode, to_g_node_alias=self.atn_alias
        )
        # Report to DispatchContract with heartbeat.algo.audit
        ptxn = PaymentTxn(self.acct.addr, self.sp, self.dc_client.app_addr, 1000)
        self.dc_client.call(
            DispatchContract.heartbeat_algo_audit,
            signed_proof=TransactionWithSigner(ptxn, self.acct.as_signer()),
            heartbeat=pong.as_dict(),
        )

    def heartbeat_from_super(self, from_alias: str, ping: HeartbeatA) -> None:
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex=ping.MyHex
        ).tuple

        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )

        LOGGER.debug(
            f"[{self.alias}] Sent HB: SuHex {pong.YourLastHex}, AtnHex {pong.MyHex}"
        )

    def timestep_from_timecoordinator(self, payload: SimTimestep):
        if self._time == 0:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.info(f"TIME STARTED: {self.time_str()}")
        elif self._time < payload.TimeUnixS:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.debug(f"Time is now {self.time_str()}")
        elif self._time == payload.TimeUnixS:
            self.repeat_timestep(payload)

    ########################################################
    #  Related to Dispatch Contract and Scada Cert
    ########################################################

    def algo_init_check(self):
        """Upon startup, see if the dispatch contract has already been created
        and also make sure Scada is appropriately funded"""

        # First, request cert transfer. This is idempotent - if scada
        # addr already owns the ScadaCertificate it does nothing

        self.request_cert_transfer()
        apps = self.client.account_info(self.acct.addr)["created-apps"]
        balances = algo_utils.get_balances(self.client, self.acct.addr)
        micro_algos = balances[0]
        if micro_algos < 100_000:
            raise Exception(f"Insufficient funds to even opt into SCADA cert!")
        if len(apps) > 1:
            raise Exception("Should only be part of one app")
        if not self.dispatch_contract_created():
            # Create, fund and do bootstrap1 of the Dispatch contract
            dc_algos = DISPATCH_CONTRACT_REPORTING_ALGOS + (
                DispatchContract.min_balance / 1_000_000
            )
            if micro_algos < (dc_algos + 2) * 1_000_000:
                raise Exception(f"Insufficient funding! Need at least {dc_algos + 2}")
            self.dc_client = ApplicationClient(
                self.client, DispatchContract(), signer=self.acct.as_signer()
            )
            app_id, app_address, transaction_id = self.dc_client.create()
            self.dc_app_id = app_id
            LOGGER.info(f"Created Dispatch Contract, app id {self.dc_app_id}")
            sp = self.dc_client.get_suggested_params()
            sp.flat_fee = True
            sp.fee = 2000
            ptxn = PaymentTxn(
                self.acct.addr,
                sp,
                self.dc_client.app_addr,
                DispatchContract.min_balance,
            )

            # Now do the first half of bootstrapping the contract by providing it
            # with enough money to fund about 32kB in boxes, along with
            # our ScadaCertId (which can be shown to match our address in the sig)

            result = self.dc_client.call(
                DispatchContract.bootstrap1,
                scada_seed=TransactionWithSigner(ptxn, self.acct.as_signer()),
                ScadaCert=self.cert_id.Idx,
            )
            LOGGER.info("Called bootstrap1 method of DispatchContract")
            LOGGER.info(
                f"Providing the scada certificate, and funding with {dc_algos} Algos"
            )
            if result.return_value != self.ta_alias:
                raise Exception(
                    f"Expected {self.ta_alias} back from bootstrapping app {self.dc_app_id}, but"
                    f" got {result.return_value}"
                )
        else:
            created_apps = self.client.account_info(self.acct.addr)["created-apps"]
            self.dc_app_id = created_apps[0]["id"]
            self.dc_client = ApplicationClient(
                client=self.client,
                app=DispatchContract(),
                signer=self.acct.as_signer(),
                app_id=self.dc_app_id,
            )
            LOGGER.info(f"Dispatch contract already exists, app id {self.dc_app_id}")
        self.in_dispatch_contract()

    def request_cert_transfer(self) -> None:
        """
        Opts into certificate and then request transfer of the ScadaCert from the GNodeFactory

        Raises:
            Exception if the scada certificate does not exist, or transfer is not successful
        """
        cert_type = AlgoCertType(self.settings.cert_type_value)
        if cert_type == AlgoCertType.SmartSig:
            raise NotImplementedError("Not prepared for SmartSig Certificates yet")

        # Look for cert id
        gnf = self.settings.public.gnf_admin_addr
        a = self.client.account_info(gnf)["created-assets"]
        scada_nft_ids = list(
            filter(
                lambda x: x["params"]["unit-name"] == "SCADA"
                and x["params"]["name"] == self.ta_alias,
                a,
            )
        )
        if len(scada_nft_ids) == 0:
            raise Exception("Scada certificate does not exist")

        self.cert_id = GwCertId_Maker(
            type=AlgoCertType.ASA, idx=scada_nft_ids[0]["index"], addr=None
        ).tuple

        balances = algo_utils.get_balances(self.client, self.acct.addr)

        # First, opt into sig
        txn = AssetOptInTxn(self.acct.addr, self.sp, self.cert_id.Idx)

        signed_txn = txn.sign(self.acct.sk)
        if self.cert_id.Idx not in balances.keys():
            try:
                self.client.send_transaction(signed_txn)
            except:
                raise Exception(
                    "Failure sending transaction on Algo blockchain. Check sandbox and funding"
                )
            algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())

        balances = algo_utils.get_balances(self.client, self.acct.addr)

        if balances[self.cert_id.Idx] == 1:
            # already own the cert
            return
        payload = ScadaCertTransfer_Maker(
            ta_alias="d1.isone.ver.keene.holly.ta",  # REPLACE WITH api_util.get_alias_from_scada_cert
            signed_proof=encoding.msgpack_encode(signed_txn),
        ).tuple

        api_endpoint = f"{self.settings.public.gnf_api_root}/scada-cert-transfer/"
        try:
            r = requests.post(url=api_endpoint, json=payload.as_dict())
        except:
            raise Exception(f"Post to {api_endpoint} failed")
        if r.status_code > 200:
            if r.status_code == 422:
                note = f"Error entering SLA: " + r.json()["detail"]
            else:
                note = r.reason
            raise Exception(f"Post to {api_endpoint} failed: {note} ")

        # check that scada now owns cert
        balances = algo_utils.get_balances(self.client, self.acct.addr)
        if balances[self.cert_id.Idx] != 1:
            raise Exception(f"Gnf claimed successful transfer but I do not own cert")

        LOGGER.info(f"Successfully received Scada Cert ASA {self.cert_id.Idx}")

    def initialize_dispatch_contract(self):
        """Starts up Dispatch contract to be used with AtomicTNode.


        Raises exception if Scada Addr
            - scada cert from settings matches what is in my acct
            - already has created an app
            - does not already have a unique Scada Cert ASA
        """

        if self.in_dispatch_contract():
            LOGGER.warning(
                f"Already in dispatch contract. Ignoring initialize_dispatch_contract!"
            )
            return

        balances = algo_utils.get_balances(self.client, self.acct.addr)
        non_algo_asas = list(set(balances.keys()) - {0})
        scada_certs = list(
            filter(
                lambda x: self.client.asset_info(x)["params"]["unit-name"] == "SCADA",
                non_algo_asas,
            )
        )

        if len(scada_certs) > 1:
            raise Exception(
                f"Scada {self.alias} has 2 scada certs. This should not happen!"
            )
        if len(scada_certs) == 0:
            raise Exception(
                "Scada acct needs to be created and funded by TaOwner, and Scada GNode "
                "needs to be authorized (with GNodeFactory sending Scada cert) "
                f"before starting up {self.alias}"
            )

        if self.cert_id.Idx != scada_certs[0]:
            raise Exception(
                f"Scada cert in settings {self.cert_id.Idx} does not match"
                f"scada cert in my acct {scada_certs[0]}"
            )

        txn = PaymentTxn(
            sender=self.acct.addr,
            receiver=self.dc_client.app_addr,
            sp=self.sp,
            amt=1000,
        )
        signed_txn = txn.sign(self.acct.sk)
        # Don't need to actually send it in order to establish signature
        payload = JoinDispatchContract_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            dispatch_contract_app_id=self.dc_app_id,
            signed_proof=encoding.msgpack_encode(signed_txn),
        ).tuple

        self.send_message(
            payload=payload,
            to_role=GNodeRole.AtomicTNode,
            to_g_node_alias=self.atn_alias,
        )
        LOGGER.info(f"Sent d = {payload.as_dict()}")

    def dispatch_contract_confirmed_received(self, payload: DispatchContractConfirmed):
        """Check the state of the app to confirm that the Atn has finished the bootstrap.
        Then
           - set the atn_gni_id to GNodeInstanceId from the payload (this is how the Scada
           will confirm identity of the AtomicTNode
           - set the atn_params to those shared by the AtomicTNode
           - opt into the Dispatch Contract (now the DispatchContract is live)"""
        if self.dc_app_id is None:
            raise Exception(
                "Thats odd ... dont have a Dispatch Contract App Id and I need to be"
                "the creator"
            )
        if self.in_dispatch_contract():
            LOGGER.warning(
                f"Ignoring DispatchContractConfirmed - already in dispatch contract {self.dc_app_id}"
            )
            return

        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            LOGGER.warning(f"Atn bootstrap is not finished. Ignoring")
            return

        self.atn_gni_id = payload.FromGNodeInstanceId
        self.atn_params = payload.AtnParams
        with open(".env", "a") as file:
            file.write(f'SCADA_ATN_GNI_ID="{payload.FromGNodeInstanceId}"\n')
        self.dc_client.opt_in()
        LOGGER.info(f"Dispatch Contract {self.dc_app_id} is live")

    #########################################################
    # Make the below into abstractmethods if pulling out base class
    #########################################################

    def sim_scada_driver_report_received(self, payload: SimScadaDriverReport) -> None:
        """This gets received right before the top of the hour, from our
        best simulation of the TerminalAsset (which is happening in the
        AtomicTNode)."""
        if payload.FromGNodeInstanceId != self.atn_gni_id:
            LOGGER.info(f"Igoring {payload} - incorrect GNodeInstanceId")
        self.boost_power_kw = payload.BoostPowerKwTimes1000 / 1000
        self.heatpump_power_kw = payload.HeatpumpPowerKwTimes1000 / 1000
        self.cop = payload.CopTimes10 / 10
        self.store_kwh = payload.StoreKwh
        self.max_store_kwh = payload.MaxStoreKwh
        self.send_snapshot()

    def send_snapshot(self):
        """Send a snapshot of current core sensed values to AtomicTNode.
        This is done every hour, and also on sensed power change."""
        if self.atn_params is None:
            return
        report_payload = SnapshotHeatpumpwithbooststore(
            FromGNodeAlias=self.alias,
            FromGNodeInstanceId=self.g_node_instance_id,
            BoostPowerKwTimes1000=int(1000 * self.boost_power_kw),
            HeatpumpPowerKwTimes1000=int(1000 * self.heatpump_power_kw),
            CopTimes10=int(10 * self.cop),
            MaxStoreKwh=get_max_store_kwh_th(self.atn_params),
            AboutTerminalAssetAlias=self.ta_alias,
        )
        self.send_message(
            payload=report_payload,
            to_role=GNodeRole.AtomicTNode,
            to_g_node_alias=self.atn_alias,
        )

    def dispatch_received(self, payload: GtDispatchBoolean) -> None:
        """
        Dispatch received from AtomicTNode

          - Checks that the GNodeAlias and GNodeInstanceId belong to its
        AtomicTNode and that we have a dispatch contract
          - Sets talking_with to true
          - Follows instructions (turns on or turns off). Will turn on or
          off boost unless AboutNodeName is "a.heatpump"

        For hourly sim:
          - Updatespower
          - Send status to AtomicTNode
        """
        if self.atn_params is None or self.in_dispatch_contract() is False:
            LOGGER.info("Igoring dispatch command, DispatchContract is not started")
        if payload.FromGNodeInstanceId != self.atn_gni_id:
            LOGGER.info(f"Igoring {payload}, not my Atn's GNodeInstanceId")
        self.talking_with = True
        if payload.AboutNodeName == "a.heatpump":
            # Making the grossly simplifying assumption that the heat pump turns on immediately
            if payload.RelayState == 1:
                new_heatpump_power_kw = self.atn_params.RatedHeatpumpElectricityKw
            else:
                new_heatpump_power_kw = 0
            if self.heatpump_power_kw != new_heatpump_power_kw:
                self.heatpump_power_kw = new_heatpump_power_kw
                self.send_snapshot()
        else:
            if payload.RelayState == 1:
                new_boost_power_kw = self.atn_params.StoreMaxPowerKw
            else:
                new_boost_power_kw = 0
            if self.boost_power_kw != new_boost_power_kw:
                self.boost_power_kw = new_boost_power_kw
                self.send_snapshot()

    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("New timestep")
        if not self.in_dispatch_contract():
            self.initialize_dispatch_contract()

    def repeat_timestep(self, payload: SimTimestep) -> None:
        LOGGER.info("Timestep received again")

    def time_str(self) -> str:
        return pendulum.from_timestamp(self.time()).strftime("%m/%d/%Y, %H:%M")

    def dispatch_contract_created(self) -> bool:
        """The SCADA only creates one kind of smart contract, which is its dispatch contract.
        So this just checks for a created app
        """
        created_apps = self.client.account_info(self.acct.addr)["created-apps"]
        if len(created_apps) > 1:
            raise NotImplementedError(
                f"SCADA not designed yet to create multiple dispatch contracts"
            )
        if len(created_apps) > 0:
            return True
        return False

    def in_dispatch_contract(self) -> bool:
        """Checks that bootstrap 2 was completed w addition of ta_trading_rights_idx,
        and also that Scada is opted in"""
        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            LOGGER.info(
                "Not in Dispatch Contract because ta_trading_rights_idx isn't in app_state"
            )
            LOGGER.info(
                "AtomicTNode has not finished its part in bootstrapping the contract"
            )
            return False
        # Todo: also check that atn is opted in
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        app_ids = list(map(lambda x: x["id"], apps))
        if self.dc_client.app_id in app_ids:
            return True
        LOGGER.info(
            "Not in Dispatch Contract because scada has not yet opted into contract"
        )
        return False



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MakeModel(StrEnum):
    """
    Determines Make/Model of device associated to a Spaceheat Node supervised by SCADA

    Choices and descriptions:

      * Freedom__Relay:
      * OmegaEzo__Ftb800Flo:
      * UnknownMake__UnknownModel:
      * GridWorks__SimBool30AmpRelay:
      * Gridworks__SimPm1:
      * GridWorks__WaterTempHighPrecision:
      * NCD__PR8-14-SPST:
      * Adafruit__642:
      * SchneiderElectric__Iem3455:
      * OpenEnergy__EmonPi:
      * Egauge__3010:
      * Rheem__XE50T10H45U0:
      * Magnelab__SCT-0300-050:
      * YMDC__SCT013-100:
      * G1__NCD_ADS1115__TEWA_NTC_10K_A: 4.6 kOhm 5% accuracy voltage divider resistor, NCD ADS1115 i2c 16 bit resolution chip, Tewa TT0P-10KC3-T105 -1500 pipe measurement NTC 10K thermistor.
      * G1__NCD_ADS1115__AMPH_NTC_10K_A: 4.6 kOhm 5% accuracy voltage divider resistor, NCD ADS1115 i2c 16 bit resolution chip, Amphenol MA100GG103BN thermistor
    """

    FREEDOM__RELAY = auto()
    OMEGAEZO__FTB800FLO = auto()
    UNKNOWNMAKE__UNKNOWNMODEL = auto()
    GRIDWORKS__SIMBOOL30AMPRELAY = auto()
    MAGNELAB__SCT0300050 = auto()
    YMDC__SCT013100 = auto()
    G1__NCD_ADS1115__TEWA_NTC_10K_A = auto()
    G1__NCD_ADS1115__AMPH_NTC_10K_A = auto()
    GRIDWORKS__SIMPM1 = auto()
    GRIDWORKS__WATERTEMPHIGHPRECISION = auto()
    NCD__PR814SPST = auto()
    ADAFRUIT__642 = auto()
    SCHNEIDERELECTRIC__IEM3455 = auto()
    OPENENERGY__EMONPI = auto()
    EGAUGE__3010 = auto()
    RHEEM__XE50T10H45U0 = auto()

    @classmethod
    def default(cls) -> "MakeModel":
        """
        Returns default value UNKNOWNMAKE__UNKNOWNMODEL
        """
        return cls.UNKNOWNMAKE__UNKNOWNMODEL

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class StrategyName(StrEnum):
    """
    Used to assign code to run a particular GNodeInstance

    Choices and descriptions:

      * NoActor: Assigned to GNodes that do not have actors
      * WorldA: Basic GridWorks world strategy
      * SupervisorA: Supervisor A Strategy, in [gridworks-atn package](https://pypi.org/project/gridworks-atn/)
      * AtnHeatPumpWithBoostStore: GridWorks strategy for an AtomicTNode representing a type of heat pump storage heating system
      * TcGlobalA: Basic GridWorks TimeCoordinator strategy, in [gridworks-timecoordinator repo](https://github.com/thegridelectric/gridworks-timecoordinator)
      * MarketMakerA: Simple GridWorks MarketMaker strategy, in [gridworks-marketmaker repo](https://github.com/thegridelectric/gridworks-marketmaker)
    """

    NoActor = auto()
    WorldA = auto()
    SupervisorA = auto()
    AtnHeatPumpWithBoostStore = auto()
    TcGlobalA = auto()
    MarketMakerA = auto()

    @classmethod
    def default(cls) -> "StrategyName":
        """
        Returns default value NoActor
        """
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type new.tadeed.send, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class NewTadeedSend(BaseModel):
    """ """

    NewTaDeedIdx: int = Field(
        title="NewTaDeedIdx",
    )
    OldTaDeedIdx: int = Field(
        title="OldTaDeedIdx",
    )
    TaDaemonAddr: str = Field(
        title="TaDaemonAddr",
    )
    ValidatorAddr: str = Field(
        title="ValidatorAddr",
    )
    SignedTadeedOptinTxn: str = Field(
        title="SignedTadeedOptinTxn",
    )
    TypeName: Literal["new.tadeed.send"] = "new.tadeed.send"
    Version: str = "000"

    @validator("TaDaemonAddr")
    def _check_ta_daemon_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"TaDaemonAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("ValidatorAddr")
    def _check_validator_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("SignedTadeedOptinTxn")
    def _check_signed_tadeed_optin_txn(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedTadeedOptinTxn failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class NewTadeedSend_Maker:
    type_name = "new.tadeed.send"
    version = "000"

    def __init__(
        self,
        new_ta_deed_idx: int,
        old_ta_deed_idx: int,
        ta_daemon_addr: str,
        validator_addr: str,
        signed_tadeed_optin_txn: str,
    ):
        self.tuple = NewTadeedSend(
            NewTaDeedIdx=new_ta_deed_idx,
            OldTaDeedIdx=old_ta_deed_idx,
            TaDaemonAddr=ta_daemon_addr,
            ValidatorAddr=validator_addr,
            SignedTadeedOptinTxn=signed_tadeed_optin_txn,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: NewTadeedSend) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> NewTadeedSend:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> NewTadeedSend:
        d2 = dict(d)
        if "NewTaDeedIdx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing NewTaDeedIdx")
        if "OldTaDeedIdx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing OldTaDeedIdx")
        if "TaDaemonAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaDaemonAddr")
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "SignedTadeedOptinTxn" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedTadeedOptinTxn")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return NewTadeedSend(
            NewTaDeedIdx=d2["NewTaDeedIdx"],
            OldTaDeedIdx=d2["OldTaDeedIdx"],
            TaDaemonAddr=d2["TaDaemonAddr"],
            ValidatorAddr=d2["ValidatorAddr"],
            SignedTadeedOptinTxn=d2["SignedTadeedOptinTxn"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from gwproto import CallableDecoder
from gwproto import Decoders
from gwproto import MQTTCodec
from gwproto import create_message_payload_discriminator
from gwproto.gs import GsPwr_Maker
from gwproto.gt.gt_dispatch_boolean import GtDispatchBoolean_Maker
from gwproto.gt.gt_sh_cli_atn_cmd import GtShCliAtnCmd_Maker
from gwproto.gt.gt_sh_status import GtShStatus_Maker
from gwproto.gt.snapshot_spaceheat import SnapshotSpaceheat_Maker


S3MessageDecoder = create_message_payload_discriminator(
    "S3MessageDecoder",
    [
        "gwproto.messages",
    ],
)


class S3MQTTCodec(MQTTCodec):
    def __init__(self):
        super().__init__(
            Decoders.from_objects(
                [
                    GtDispatchBoolean_Maker,
                    GtShCliAtnCmd_Maker,
                    GtShStatus_Maker,
                    SnapshotSpaceheat_Maker,
                ],
                message_payload_discriminator=S3MessageDecoder,
            ).add_decoder(
                "p", CallableDecoder(lambda decoded: GsPwr_Maker(decoded[0]).tuple)
            )
        )

    def validate_source_alias(self, source_alias: str):
        pass



import json
import typing
from typing import Any
from typing import List
from typing import Optional

import boto3
import pendulum
from gwproto import Message
from gwproto.messages import GtDispatchBoolean
from gwproto.messages import GtDispatchBoolean_Maker
from gwproto.messages import GtShStatus
from gwproto.messages import GtShStatus_Maker
from gwproto.messages import GtShStatusEvent
from gwproto.messages import GtTelemetry
from gwproto.messages import GtTelemetry_Maker
from gwproto.messages import SnapshotSpaceheat
from gwproto.messages import SnapshotSpaceheat_Maker
from gwproto.messages import SnapshotSpaceheatEvent
from pydantic import BaseModel

from gwatn.api_types_hack import HackTypeMakerByName
from gwatn.csv_makers.codec import S3MQTTCodec


OUT_STUB = "output_data/scada_report_a"

DOWNLOADED_FILE_TYPES = [
    GtDispatchBoolean_Maker.type_alias,
    GtShStatus_Maker.type_alias,
    SnapshotSpaceheat_Maker.type_alias,
    GtTelemetry_Maker.type_alias,
    SnapshotSpaceheatEvent.__fields__["TypeName"].default,
    GtShStatusEvent.__fields__["TypeName"].default,
]


class FileNameMeta(BaseModel):
    FromGNodeAlias: str
    PayloadTypeName: str
    UnixTimeMs: int
    FileName: str
    TypeName: str = "file.name.meta.000"


class StatusOutputRow(BaseModel):
    TimeUnixMs: int
    IntTimeUnixS: int
    Milliseconds: int
    TimeUtc: str
    Value: Optional[int]
    TelemetryName: Optional[str]
    AboutShNode: str
    FromShNode: Optional[str]
    DispatchCmd: Optional[int]
    TypeName: str = "ui.status.output.row.000"


class SnapshotWithSendTime(BaseModel):
    SendTimeUnixMs: int
    Snapshot: SnapshotSpaceheat
    TypeName: str = "wrapped.snapshot.spaceheat.000"


class ScadaReportA_Maker:
    def __init__(self, out_stub=OUT_STUB):
        self.s3 = boto3.client("s3")
        self.aws_bucket_name = "gwdev"
        self.world_instance_name = "hw1__1"
        self.out_stub = f"{out_stub}"
        self.output_type_name = "atn.ui.000"
        self.processed_file_name_meta_list = List[FileNameMeta]
        self.new_file_name_meta_list = List[FileNameMeta]
        self.latest_status_list: List[GtShStatus] = []
        self.latest_snapshot_list: List[SnapshotSpaceheat] = []
        self.latest_dispatch_list: List[GtDispatchBoolean] = []
        self.status_rows_to_write: List[StatusOutputRow] = []
        self.mqtt_codec = S3MQTTCodec()
        print(f"Initialized {self.__class__}")

    def on_gw_message(self, from_g_node_alias: str, payload):
        if from_g_node_alias not in self.ATN_ALIAS_LIST:
            print(f"Does not record {from_g_node_alias}.")
        else:
            if isinstance(payload, GtShStatus):
                self.latest_status_list.append(payload)

    def get_atn_alias_from_alias(self, alias: str):
        x = alias.split(".")
        if x[-1] == "scada":
            return ".".join(x[:-2])
        else:
            return alias

    def get_status_rows_from_telemetry(
        self, from_sh_node_alias: str, payload: GtTelemetry
    ) -> List[StatusOutputRow]:
        time_unix_ms = payload.ScadaReadTimeUnixMs
        time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
        row = StatusOutputRow(
            TimeUnixMs=time_unix_ms,
            IntTimeUnixS=int(time_unix_ms / 1000),
            Milliseconds=int(time_unix_ms) % 1000,
            TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
            Value=payload.Value,
            TelemetryName=payload.Name.value,
            AboutShNode=from_sh_node_alias,
            FromShNode=from_sh_node_alias,
        )
        return [row]

    def get_status_rows_from_dispatch(
        self, payload: GtDispatchBoolean
    ) -> List[StatusOutputRow]:
        rows = []
        time_unix_ms = payload.SendTimeUnixMs
        time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
        row = StatusOutputRow(
            TimeUnixMs=time_unix_ms,
            IntTimeUnixS=int(time_unix_ms / 1000),
            Milliseconds=int(time_unix_ms) % 1000,
            TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
            AboutShNode=payload.AboutNodeAlias,
            FromShNode=payload.FromGNodeAlias,
            DispatchCmd=payload.RelayState,
        )
        rows.append(row)
        return rows

    def get_status_rows_from_snapshot(
        self, payload: SnapshotWithSendTime
    ) -> List[StatusOutputRow]:
        rows = []
        from_g_node_alias = payload.Snapshot.FromGNodeAlias
        snapshot = payload.Snapshot.Snapshot
        time_unix_ms = payload.SendTimeUnixMs
        time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
        for i in range(len(snapshot.ValueList)):
            row = StatusOutputRow(
                TimeUnixMs=time_unix_ms,
                IntTimeUnixS=int(time_unix_ms / 1000),
                Milliseconds=int(time_unix_ms) % 1000,
                TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                Value=snapshot.ValueList[i],
                TelemetryName=snapshot.TelemetryNameList[i].value,
                AboutShNode=snapshot.AboutNodeAliasList[i],
                FromShNode=from_g_node_alias,
            )
            rows.append(row)
        return rows

    def get_status_rows_from_status(self, payload: GtShStatus) -> List[StatusOutputRow]:
        rows: List[StatusOutputRow] = []
        for single in payload.SimpleTelemetryList:
            for i in range(len(single.ValueList)):
                time_unix_ms = single.ReadTimeUnixMsList[i]
                time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
                row = StatusOutputRow(
                    TimeUnixMs=time_unix_ms,
                    IntTimeUnixS=int(time_unix_ms / 1000),
                    Milliseconds=int(time_unix_ms) % 1000,
                    TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                    Value=single.ValueList[i],
                    TelemetryName=single.TelemetryName.value,
                    AboutShNode=single.ShNodeAlias,
                    FromShNode=single.ShNodeAlias,
                )
                rows.append(row)

        for cmd in payload.BooleanactuatorCmdList:
            about_node_alias = cmd.ShNodeAlias
            for i in range(len(cmd.RelayStateCommandList)):
                time_unix_ms = cmd.CommandTimeUnixMsList[i]
                time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
                row = StatusOutputRow(
                    TimeUnixMs=time_unix_ms,
                    IntTimeUnixS=int(time_unix_ms / 1000),
                    Milliseconds=int(time_unix_ms) % 1000,
                    TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                    AboutShNode=about_node_alias,
                    DispatchCmd=cmd.RelayStateCommandList[i],
                )
                rows.append(row)

        for multi in payload.MultipurposeTelemetryList:
            for i in range(len(multi.ValueList)):
                time_unix_ms = multi.ReadTimeUnixMsList[i]
                time_utc = pendulum.from_timestamp(time_unix_ms / 1000)
                row = StatusOutputRow(
                    TimeUnixMs=time_unix_ms,
                    IntTimeUnixS=int(time_unix_ms / 1000),
                    Milliseconds=int(time_unix_ms) % 1000,
                    TimeUtc=time_utc.strftime("%Y-%m-%d %H:%M:%S"),
                    Value=multi.ValueList[i],
                    TelemetryName=multi.TelemetryName.value,
                    AboutShNode=multi.AboutNodeAlias,
                    FromShNode=multi.SensorNodeAlias,
                )
                rows.append(row)
        return rows

    def get_message_from_s3(self, file_name_meta: FileNameMeta) -> Message:
        s3_object = self.s3.get_object(
            Bucket=self.aws_bucket_name, Key=file_name_meta.FileName
        )
        message = typing.cast(
            Message, self.mqtt_codec.decode("gw", s3_object["Body"].read())
        )
        match message.Payload:
            case SnapshotSpaceheat():
                message.Payload = SnapshotWithSendTime(
                    SendTimeUnixMs=file_name_meta.UnixTimeMs, Snapshot=message.Payload
                )
            case SnapshotSpaceheatEvent():
                message.Payload = SnapshotWithSendTime(
                    SendTimeUnixMs=file_name_meta.UnixTimeMs,
                    Snapshot=message.Payload.snap,
                )
        return message

    def has_this_days_folder(self, time_s: int) -> bool:
        d = pendulum.from_timestamp(time_s)
        this_days_folder_name = d.strftime("%Y%m%d")
        prefix = f"{self.world_instance_name}/eventstore/{this_days_folder_name}"

        r = self.s3.list_objects_v2(Bucket=self.aws_bucket_name, Prefix=prefix)
        if "Contents" in r.keys():
            return True
        return False

    def get_date_folder_list(
        self, start_time_unix_ms: int, duration_hrs: int
    ) -> List[str]:
        start_s = start_time_unix_ms / 1000
        folder_list: List[str] = []
        found_latest_earlier: bool = False
        i = 0
        while (not found_latest_earlier) and i < 5:
            t = start_s - 3600 * 24 * (i + 1)
            if self.has_this_days_folder(t):
                folder_list.append(pendulum.from_timestamp(t).strftime("%Y%m%d"))
                found_latest_earlier = True
            i += 1

        if self.has_this_days_folder(start_s):
            folder_list.append(pendulum.from_timestamp(start_s).strftime("%Y%m%d"))

        add_hrs = 0
        while add_hrs < duration_hrs:
            add_hrs += 24
            add_hrs = min(add_hrs, duration_hrs)
            t = start_s + add_hrs * 3600
            if self.has_this_days_folder(t):
                folder_list.append(pendulum.from_timestamp(t).strftime("%Y%m%d"))

        return list(set(folder_list))

    def get_file_name_meta_list(
        self,
        start_time_unix_ms: int,
        end_time_unix_ms: int,
        date_folder_list: List[str],
        g_node_alias_list: List[str],
        type_name_list: Optional[List[str]] = None,
    ):
        if type_name_list is None:
            type_name_list = DOWNLOADED_FILE_TYPES
        fn_list: List[FileNameMeta] = []
        for date_folder in date_folder_list:
            prefix = f"{self.world_instance_name}/eventstore/{date_folder}/"
            paginator = self.s3.get_paginator("list_objects_v2")
            pages = paginator.paginate(Bucket=self.aws_bucket_name, Prefix=prefix)
            file_name_list = []
            for page in pages:
                for obj in page["Contents"]:
                    file_name_list.append(obj["Key"])

            for file_name in file_name_list:
                try:
                    from_g_node_alias = file_name.split("/")[-1].split("-")[0]
                    payload_type_name = file_name.split("/")[-1].split("-")[1]
                    payload_unix_time_ms = int(file_name.split("/")[-1].split("-")[2])
                except:
                    raise Exception(f"Failed file name parsing with {file_name}")

                if (
                    from_g_node_alias in g_node_alias_list
                    and payload_type_name in type_name_list
                    and payload_unix_time_ms > start_time_unix_ms - 300_000
                    and payload_unix_time_ms < end_time_unix_ms
                ):
                    fn_list.append(
                        FileNameMeta(
                            FromGNodeAlias=from_g_node_alias,
                            PayloadTypeName=payload_type_name,
                            UnixTimeMs=payload_unix_time_ms,
                            FileName=file_name,
                        )
                    )

        return fn_list

    def get_csv_rows(
        self, start_time_unix_ms: int, duration_hrs: int, atn_alias: str
    ) -> List[StatusOutputRow]:
        g_node_alias_list = [atn_alias, atn_alias + ".scada", atn_alias + ".scada"]
        start_time_utc = pendulum.from_timestamp(start_time_unix_ms / 1000)
        end_time_utc = start_time_utc + pendulum.duration(hours=duration_hrs)
        end_time_unix_ms = end_time_utc.int_timestamp * 1000
        date_folder_list = self.get_date_folder_list(start_time_unix_ms, duration_hrs)
        fn_list = self.get_file_name_meta_list(
            start_time_unix_ms=start_time_unix_ms,
            end_time_unix_ms=end_time_unix_ms,
            date_folder_list=date_folder_list,
            g_node_alias_list=g_node_alias_list,
        )

        rows: List[StatusOutputRow] = []
        for i in range(len(fn_list)):
            fn = fn_list[i]
            message = self.get_message_from_s3(fn)
            match message.Payload:
                case GtDispatchBoolean():
                    rows += self.get_status_rows_from_dispatch(message.Payload)
                case GtShStatus():
                    rows += self.get_status_rows_from_status(message.Payload)
                case GtShStatusEvent():
                    rows += self.get_status_rows_from_status(message.Payload.status)
                case SnapshotWithSendTime():
                    rows += self.get_status_rows_from_snapshot(message.Payload)
        rows = sorted(rows, key=lambda x: x.TimeUnixMs)
        return rows

    def make_csv(
        self,
        start_s,
        duration_hrs: int = 48,
        atn_alias: str = "hw1.isone.ct.newhaven.orange1",
    ):
        s = start_s - (start_s % 3600)
        start_time_utc = pendulum.from_timestamp(s)
        print(f"starting at {start_time_utc.strftime('%Y-%m-%d %H:%M')}")
        start_time_unix_ms = s * 1000

        rows = self.get_csv_rows(
            start_time_unix_ms=start_time_unix_ms,
            duration_hrs=duration_hrs,
            atn_alias=atn_alias,
        )
        lines = [
            "TimeUtc, TimeUnixS, Ms, Value, TelemetryName, AboutShNode, FromShNode, DispatchCmd\n"
        ]
        for row in rows:
            line = f"{row.TimeUtc}, {row.IntTimeUnixS}, {row.Milliseconds}"
            if row.Value is None:
                line += ", "
            else:
                line += f",{row.Value}"
            if row.TelemetryName is None:
                line += ", "
            else:
                line += f", {row.TelemetryName}"
            line += f", {row.AboutShNode}"
            if row.FromShNode is None:
                line += ", "
            else:
                line += f", {row.FromShNode}"
            if row.DispatchCmd is None:
                line += ", \n"
            else:
                line += f", {row.DispatchCmd} \n"
            lines.append(line)

        atn_end = atn_alias.split(".")[-1]
        file_name = f"{self.out_stub}/{atn_end}-{start_time_utc.strftime('%Y%m%d-%H%M')}-{duration_hrs}-{atn_alias}.csv"
        print(f"Writing {file_name}")
        with open(file_name, "w") as outfile:
            outfile.writelines(lines)



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MessageCategorySymbol(StrEnum):
    """
    Shorthand symbols for MessageCategory000 Enum, used in meta-data like routing keys

    Choices and descriptions:

      * unknown: Default value
      * rj: Serialized Json message sent on the world rabbit broker from one GNode actor to another
      * rjb: Serailized Json message broadcast on the world rabbit broker by a GNode actor
      * s: GwSerial protocol message sent on the world rabbit broker
      * gw: Serialized Json message following MQTT topic format, sent on the world rabbit broker
      * post: REST API post
      * postack: REST API post response
      * get: REST API GET
    """

    unknown = auto()
    rj = auto()
    rjb = auto()
    s = auto()
    gw = auto()
    post = auto()
    postack = auto()
    get = auto()

    @classmethod
    def default(cls) -> "MessageCategorySymbol":
        """
        Returns default value unknown
        """
        return cls.unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



import json
import logging
import subprocess
import sys
import time
from typing import List

import gridworks.api_utils as api_utils
import gridworks.dev_utils.algo_setup as algo_setup
from gridworks.algo_utils import BasicAccount
from gridworks.utils import RestfulResponse
from pydantic import SecretStr
from rich.pretty import pprint

import gwatn.config as config
from gwatn.dev_utils import DevDiscoverer
from gwatn.dev_utils import DevTaOwner
from gwatn.dev_utils import DevValidator
from gwatn.types import BaseGNodeGt_Maker
from gwatn.types import TadeedSpecsHack_Maker


logging.basicConfig(level="INFO")
LOGGER = logging.getLogger(__name__)


def certify_molly_metermaid() -> RestfulResponse:
    molly = DevValidator(config.ValidatorSettings())
    cert_idx = api_utils.get_validator_cert_idx(validator_addr=molly.acct.addr)
    if cert_idx is not None:
        raise Exception(
            f"There is already a Validator Certificate for Molly! Please ./sandbox reset and start the demo over."
        )
    return molly.post_tavalidatorcert_algo_create()


def create_ta_owner(plant: str) -> DevTaOwner:
    file = f"input_data/eventstore/d1.isone.ver.keene.{plant}.ta-tadeed.specs.hack.json"
    with open(file) as f:
        payload = TadeedSpecsHack_Maker.dict_to_tuple(json.load(f))

    settings = config.TaOwnerSettings()
    ta_owner_acct = BasicAccount()
    settings.sk = SecretStr(ta_owner_acct.sk)
    settings.initial_ta_alias = payload.TerminalAssetAlias
    settings.ta_daemon_api_port = payload.DaemonPort
    ta_owner = DevTaOwner(settings)
    return ta_owner


def create_ta_owners(plants: List[str]) -> List[DevTaOwner]:
    owners: List[DevTaOwner] = []
    for plant in plants:
        owners.append(create_ta_owner(plant))
    return owners


def start_ta_owners(ta_owners: List[DevTaOwner]) -> RestfulResponse:
    started_ta_owners: List[DevTaOwner] = []
    for owner in ta_owners:
        try:
            owner.start()
        except:
            for owner in started_ta_owners:
                owner.stop()
            return RestfulResponse(
                Note=f"Error starting {owner.settings.initial_ta_alias} owner.",
                HttpStatusCode=422,
            )
        started_ta_owners.append(owner)
    return RestfulResponse(Note="Successfully started all TaOwners and TaDaemons")


def create_terminal_asset(ta_owner: DevTaOwner) -> RestfulResponse:
    molly = DevValidator(config.ValidatorSettings())
    rr = ta_owner.request_ta_certification()
    if rr.HttpStatusCode > 200:
        note = (
            f"Stopping demo due to errors in requesting ta certification for "
            f"{ta_owner.settings.initial_ta_alias}" + rr.Note
        )
        return RestfulResponse(Note=note, HttpStatusCode=422)
    pprint(rr)
    terminal_asset = BaseGNodeGt_Maker.dict_to_tuple(rr.PayloadAsDict)
    ta_deed_idx = terminal_asset.OwnershipDeedId
    LOGGER.info(f"Made TaDeed {ta_deed_idx} for {terminal_asset.Alias}")

    rr = molly.certify_terminal_asset(
        ta_deed_idx=ta_deed_idx,
        ta_daemon_addr=ta_owner.ta_daemon_acct.addr,
        ta_owner_addr=ta_owner.acct.addr,
        micro_lat=ta_owner.settings.micro_lat,
        micro_lon=ta_owner.settings.micro_lon,
    )

    if rr.HttpStatusCode > 200:
        note = (
            "Stopping demo due to errors certifying terminal asset for "
            f"{ta_owner.settings.initial_ta_alias}" + rr.Note
        )
        return RestfulResponse(Note=note, HttpStatusCode=422)
    pprint(rr)
    return rr


def create_terminal_assets(ta_owners: List[DevTaOwner]) -> RestfulResponse:
    for ta_owner in ta_owners:
        if not isinstance(ta_owner, DevTaOwner):
            return RestfulResponse(
                Note=f"{ta_owner} is not a DevTaOwner!", HttpStatusCode=422
            )
    for ta_owner in ta_owners:
        rr = create_terminal_asset(ta_owner)
        if rr.HttpStatusCode > 200:
            for ta_owner in ta_owners:
                ta_owner.stop()
            return rr
    return RestfulResponse(Note="Success with create_terminal_assets")


def create_scadas(ta_owners: List[DevTaOwner]) -> RestfulResponse:
    for ta_owner in ta_owners:
        if not isinstance(ta_owner, DevTaOwner):
            return RestfulResponse(
                Note=f"{ta_owner} is not a DevTaOwner!", HttpStatusCode=422
            )
    for ta_owner in ta_owners:
        rr = ta_owner.create_scada_g_node()
        if rr.HttpStatusCode > 200:
            for ta_owner in ta_owners:
                ta_owner.stop()
            return rr
    return RestfulResponse(Note="Success with create_scadas")


def enter_slas(ta_owners: List[DevTaOwner]) -> RestfulResponse:
    for ta_owner in ta_owners:
        if not isinstance(ta_owner, DevTaOwner):
            return RestfulResponse(
                Note=f"{ta_owner} is not a DevTaOwner!", HttpStatusCode=422
            )
    for ta_owner in ta_owners:
        rr = ta_owner.enter_sla()
        if rr.HttpStatusCode > 200:
            for ta_owner in ta_owners:
                ta_owner.stop()
            return rr
    return RestfulResponse(Note="Success with entering ServiceLevelAgreements")


def create_new_ctn():
    ada = DevDiscoverer(settings=config.DiscovererSettings())
    rr = ada.post_discoverycert_algo_create()
    LOGGER.info("Ada received response to discoverycert algo")


demo_plant_names: List[str] = [
    "holly",
    "juniper",
    "violet",
    "lettuce",
    "mushroom",
    "nettle",
    "orange",
    "pine",
    "quinoi",
    "rose",
    "leek",
    "umbrella",
    "kale",
    "wasabi",
    "yarrow",
    "zinnia",
    "apple",
    "beet",
    "acacia",
    "buxus",
    "calathea",
    "chestnut",
    "columnea",
    "cuphea",
    "dracaena",
    "eucharis",
    "gardenia",
    "hedera",
    "ixora",
    "lemon",
    "mansoa",
    "murraya",
    "pepper",
    "pilea",
    "redwood",
    "stapelia",
    "acca",
    "begonia",
    "callisia",
    "citronella",
    "cordyline",
    "cyanotis",
    "drosera",
    "euphorbia",
    "gasteria",
    "orchid",
    "jacobinia",
    "lepanthes",
    "maple",
    "myrtle",
    "pandanus",
    "punica",
    "rhapis",
    "strelitzia",
    "acorus",
    "billbergia",
    "campanula",
    "cleyera",
    "corokia",
    "cycas",
    "duchesnea",
    "exacum",
    "gloriosa",
    "hibiscus",
    "jasmine",
    "lilium",
    "maranta",
    "neoregelia",
    "passiflora",
    "pineapple",
    "rhoeo",
    "thistle",
    "aechmea",
    "biophytum",
    "capsicum",
    "clivia",
    "costus",
    "cyclamen",
    "carrot",
    "fatshedera",
    "gongora",
    "hoffmannia",
    "jatropha",
    "redbud",
    "maxillaria",
    "nepenthes",
    "celery",
    "pisonia",
    "rivina",
    "thunia",
    "aeonium",
    "birch",
    "caryota",
    "coccoloba",
    "cottonwood",
    "cyperus",
    "elm",
    "fatsia",
    "guzmania",
    "howea",
    "kalanchoe",
    "lily",
    "medinilla",
    "nerine",
    "pellionia",
    "pleione",
    "rochea",
    "tolmiea",
    "agapetes",
    "blechnum",
    "cattleya",
    "coconut",
    "cotyledon",
    "datura",
    "encyclia",
    "ficus",
    "gynura",
    "hoya",
    "kohleria",
    "liriope",
    "miltonia",
    "nertera",
    "pentas",
    "plumbago",
    "ruellia",
    "tomato",
    "agave",
    "bouvardia",
    "ceropegia",
    "coffea",
    "crassula",
    "dionaea",
    "epidendrum",
    "fir",
    "haemaria",
    "hydrangea",
    "laelia",
    "livistona",
    "mimosa",
    "oak",
    "peperomia",
    "polyscias",
    "saffron",
    "radish",
    "aloe",
    "brunfelsia",
    "chenolle",
    "coleus",
    "crinum",
    "dipladenia",
    "episcia",
    "fittonia",
    "haworth",
    "impatiens",
    "lantana",
    "macodes",
    "molineria",
    "oleander",
    "petunia",
    "primrose",
    "sanchezia",
    "willow",
    "beech",
    "cactus",
    "cherimoya",
    "colmanara",
    "crossandra",
    "dischidia",
    "erica",
    "fuchsia",
    "haworthia",
    "iresine",
    "laurus",
    "mallow",
    "monstera",
    "oncidium",
    "phoenix",
    "primula",
    "spruce",
    "camellia",
]


def cert_creation(sim_size: int) -> List[DevTaOwner]:
    plant_names = demo_plant_names[0:sim_size]
    if sim_size == 0:
        raise Exception("No simulated TerminalAssets. Stopping")
    elif sim_size == 1:
        print(f"Running simulation for 1 TerminalAsset (holly.ta)")
    elif sim_size > 4:
        raise Exception("Please choose between 1 and 4 TerminalAssets")
    else:
        print(f"Running simulation for {sim_size} TerminalAssets")
    time.sleep(2)
    print("")

    print("")
    print("")
    print("Funding GNodeFactory")

    algo_setup.dev_fund_to_min(config.Public().gnf_admin_addr, 25)
    algo_setup.dev_fund_to_min(config.Public().gnf_graveyard_addr, 1)

    print("")
    print("")
    print(f"Funding {sim_size} Atn accts")
    print("")
    print("")

    atn_address_list = [
        "4JHRDNY4F6RCVGPALZULZWZNVP3OKT3DATEOLINCGILVPGHUOFY7KCHVIQ",
        "YL3MAWSIST2DWX5GOTZYVX74GNSAVMX52RKSEDT4KIO644JCCRFTFKM5UM",
        "CWOLXCXZKLYLORBQQCI4AUHA5CLOLUXRDZEJZI4S3F6WUNVIAF4MX5EW4U",
        "R3PKD54UOAOW6MTPO7ECZ6YX4COQWN5BJM4OZIHYWFVVGAITM53RGUF6LI",
    ]

    for i in range(sim_size):
        addr = atn_address_list[i]
        algo_setup.dev_fund_to_min(addr=addr, min_algos=25)

    addr = atn_address_list[0]
    print("")
    print("")
    print("Certifying MollyMetermaid as a TaValidator")
    print("")
    print("")
    time.sleep(2)

    print("")
    print("")
    print("Starting up Molly Metermaid's Validator API")
    print("")
    print("")
    time.sleep(2)
    cmd = "docker compose -f docker-api.yml up -d"
    subprocess.run(cmd.split())

    print("Verify that it works by inspecting http://localhost:8001/docs")
    print("")
    print("")

    time.sleep(2)
    print("This also started up the API half of the MarketMaker.")
    print("Verify that it is working:")

    print("- http://localhost:7997/ shows market maker information")
    print(
        "- http://localhost:7997/get-time/, which should be 0 unix time, but will show the time"
    )
    print(" of the simulation once that starts")

    rr = certify_molly_metermaid()

    pprint(rr)
    if rr.HttpStatusCode > 200:
        raise Exception("Stopping demo due to errors")

    print("")
    print("")
    print(f"Creating {sim_size} TaOwners")
    print("")
    print("")
    time.sleep(2)
    ta_owners = create_ta_owners(plant_names)
    print("")
    print("")
    print(f"Creating {sim_size} TaDaemons")
    print("")
    print("")
    time.sleep(2)
    start_ta_owners(ta_owners)

    print("")
    print("")
    print("TaDaemons are now running, each with their own RestAPI.")
    print("")
    print("")
    time.sleep(2)
    print("Any TaDeeds they own will show up at the following endpoints:")
    print("")
    print("")
    for owner in ta_owners:
        print(
            f"Inspect {owner}'s deeds at http://localhost:{owner.settings.ta_daemon_api_port}/owned-tadeeds/"
        )

    print("")
    print("")
    time.sleep(2)
    print("They do not yet own any TaDeeds.")
    print("")
    print("")
    time.sleep(2)
    input("HIT RETURN TO CONTINUE")

    print("")
    print("")
    print(f"Creating {sim_size} TerminalAssets")
    print("")
    print("")
    time.sleep(2)

    rr = create_terminal_assets(ta_owners)

    if rr.HttpStatusCode == 200:
        print("Success!")
        print("")
        print("")
        time.sleep(2)
        print(
            "TaDaemon Algorand addresses now hold TaDeeds on behalf of their TaOwners"
        )
        print("")
        print("")
        time.sleep(2)
        print("Inspect them at:")

        for owner in ta_owners:
            print(
                f"Inspect {owner}'s deeds at http://localhost:{owner.settings.ta_daemon_api_port}/owned-tadeeds/"
            )

    else:
        for ta_owner in ta_owners:
            ta_owner.stop()  # Does the same
        raise Exception(
            f"Something went wrong creating TerminalAssets: {rr.HttpStatusCode}, {rr.Note}"
        )

    print("")
    print("")
    print(
        "The AtomicTNodes do not yet own trading rights. Inspect trading right owners at:"
    )
    for owner in ta_owners:
        print(
            f"Inspect {owner}'s deeds at http://localhost:{owner.settings.ta_daemon_api_port}/trading-rights/"
        )

    print("")
    print("")
    time.sleep(2)
    input("HIT RETURN TO CONTINUE")
    print("")
    print("")
    print(
        "In fact the AtomicTNodes do not exist yet. Go to http://d1-1.electricity.works:15672/#/queues"
    )
    print("")
    print("")
    time.sleep(1)
    print("Username and password are the same:")
    print("")
    print("smqPublic")
    print("")
    print("")
    time.sleep(1)
    print(
        "You will only see the dummy queues, the world d1-Fxxx, and the market maker ...keene.F-xxx"
    )
    # input("To start the atn actors (and time coordinator) in a docker instance, HIT RETURN")
    # )
    # print("")
    # print("")
    # print("It takes about 5 seconds for them to shop up. Look for them at")
    # print("http://d1-1.electricity.works:15672/#/queues")
    # print(
    #     "Once their queues exist they ready to enter their Service Level Agreements and get their trading rights"
    # )
    # input("HIT RETURN TO CONTINUE")
    rr = enter_slas(ta_owners)

    if rr.HttpStatusCode == 200:
        print("")
        print("")
        time.sleep(2)
        print(
            "Daemons have transferred TradingRights to their AtomicTNodes. Inspect at above pages"
        )
        print("")
        print("")
    else:
        cmd = "docker compose -f docker-actor.yml down"
        subprocess.run(cmd.split())
        raise Exception("Something went wrong entering Service Level Agreements")

    print("")
    print("")
    print(f"Getting SCADA Certs for {sim_size} Scada")
    print("")
    print("")
    time.sleep(2)

    rr = create_scadas(ta_owners)

    if rr.HttpStatusCode == 200:
        print("Success!")
        print("")
        print("")
        time.sleep(2)
        print("Scada accounts are funded, and Scada Certs are created")

    else:
        for ta_owner in ta_owners:
            ta_owner.stop()  # Does the same
        raise Exception(
            f"Something went wrong creating TerminalAssets: {rr.HttpStatusCode}, {rr.Note}"
        )
    return ta_owners



"""Type tavalidatorcert.algo.transfer, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

import gridworks.algo_utils as algo_utils
import gridworks.api_utils as api_utils
import gridworks.gw_config as config
from algosdk import encoding
from algosdk.future import transaction
from algosdk.v2client.algod import AlgodClient
from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class TavalidatorcertAlgoTransfer(BaseModel):
    """Used for Step 2 of TaValidator certification.

    Meant to be sent from a pending TaValidator  to the GNodeFactory (Gnf), so the
    Gnf will transfer its ValidatorCert to the pending TaValidator's Algorand address.
    [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html).
    """

    ValidatorAddr: str = Field(
        title="The address of the pending TaValidator",
    )
    HalfSignedCertTransferMtx: str = Field(
        title="Algo multi-transaction for certificate transfer, with 1 of 2 signatures",
    )
    TypeName: Literal["tavalidatorcert.algo.transfer"] = "tavalidatorcert.algo.transfer"
    Version: str = "000"

    @validator("ValidatorAddr")
    def check_validator_addr(cls, v: str) -> str:
        """
        Axiom 4: TaValidator has sufficient Algos.
        MultiAccount [GnfAdminAddr, ValidatorAddr] must have enough Algos to meet
        the GNodeFactory criterion.
        """
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )

        multi_addr = api_utils.get_validator_account_with_admin(v).addr
        if (
            algo_utils.algos(multi_addr)
            < config.Public().ta_validator_funding_threshold_algos
        ):
            raise ValueError(
                f"Axiom 4: 2-sig Multi [GnfAdminAddr, ValidatorAddr] insufficiently funded "
                f" with {algo_utils.algos(v)} algos. "
                f" Needs {config.Public().ta_validator_funding_threshold_algos} algos. "
            )

        return v

    @validator("HalfSignedCertTransferMtx")
    def _check_half_signed_cert_transfer_mtx(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"HalfSignedCertTransferMtx failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded HalfSignedCertTransferMtx must have type MultisigTransaction from the
        2-sig MultiAccount  [GnfAdminAddr, ValidatorAddr], signed by the ValidatorAddr.
        [More info](https://gridworks.readthedocs.io/en/latest/g-node-factory.html#gnfadminaddr)
        """
        mtx = encoding.future_msgpack_decode(v.get("HalfSignedCertTransferMtx", None))
        ValidatorAddr = v.get("ValidatorAddr")
        gnf_admin_addr = config.Public().gnf_admin_addr
        multi = algo_utils.MultisigAccount(
            version=1,
            threshold=2,
            addresses=[gnf_admin_addr, ValidatorAddr],
        )

        if not isinstance(mtx, transaction.MultisigTransaction):
            raise ValueError(
                "Axiom 1: Decoded HalfSignedCertTransferMtx must have type MultisigTransaction,"
                f" got {type(mtx)}."
            )

        if not multi.addr == mtx.multisig.address():
            raise ValueError(
                "Axiom 1: Decoded HalfSignedCertTransferMtx must come from the 2-sig MultiAccount ,"
                f" [GnfAdminAddr, ValidatorAddr], got {type(mtx)}."
            )
        if mtx.multisig.subsigs[1].signature is None:
            raise ValueError(
                "Axiom 1: Decoded HalfSignedCertTransferMtx missing TaValidator signature.}"
            )
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: Transfers correct certificate.
         - The transaction must be the transfer of an Algorand Standard Asset
         - The sender must be the 2-sig Multi [GnfAdminAddr, TaValidatorAddr], which also created and owns the ASA
         - It must be getting sent to the ValidatorAddr
         -The ASA must have:
           - Total = 1
           - UnitName=VLDITR
           - GnfAdminAddr as manage
           - AssetName not blank.
        - The transfer amount must be 1
        [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate)

        Axiom 3: TaValidator has opted in.
        ValidatorAddr must be opted into the transferring ASA.
        """
        mtx = encoding.future_msgpack_decode(v.get("HalfSignedCertTransferMtx", None))
        txn = mtx.transaction
        gnf_admin_addr = config.Public().gnf_admin_addr
        ValidatorAddr = v.get("ValidatorAddr")
        multi = algo_utils.MultisigAccount(
            version=1,
            threshold=2,
            addresses=[gnf_admin_addr, ValidatorAddr],
        )

        if not isinstance(txn, transaction.AssetTransferTxn):
            raise ValueError(
                "Axiom 2: The transaction must have type AssetTransferTxn"
                f" got {type(txn)}."
            )
        if not txn.sender == multi.addr:
            raise ValueError(
                "Axiom 2: The transfer must be sending from the 2-sig Multi [GNfAdminAddr, TaValidatorAddr]"
            )

        settings = config.VanillaSettings()
        client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        multi_balances = algo_utils.get_balances(client, multi.addr)
        if not txn.index in multi_balances.keys():
            raise ValueError(
                "Axiom 2: The 2-sig Multi [GnfAdminAddr, TaValidatorAddr] must have created and own the ASA"
            )
        if multi_balances[txn.index] != 1:
            raise ValueError(
                "Axiom 2: The ASA creator must be the 2-sig Multi [GnfAdminAddr, TaValidatorAddr]"
            )
        if not txn.receiver == ValidatorAddr:
            raise ValueError("Axiom 2: The ASA receiver must be the TaValiadatorAddr")
        if not txn.amount == 1:
            raise ValueError(f"Axiom 2: The amount sent must be 1, not {txn.amount}")
        asa_params = client.asset_info(txn.index)["params"]
        if asa_params["total"] != 1:
            raise ValueError(
                f"Axiom 2: The ASA Total must be 1, not {asa_params['total']}"
            )
        if asa_params["decimals"] != 0:
            raise ValueError(
                f"Axiom 2: The ASA Decimals must be 0, not {asa_params['decimals']}"
            )
        if asa_params["manager"] != gnf_admin_addr:
            raise ValueError(
                f"Axiom 2: The ASA manager must be GnfAdminAddr, not {asa_params['manager']}"
            )
        if asa_params["unit-name"] != "VLDTR":
            raise ValueError(
                f"Axiom 2: The ASA UnitName must be 'VLDTR, not {asa_params['unit-name']}"
            )
        if asa_params["name"] is None:
            raise ValueError(f"Axiom 2: The ASA AssetName cannot be None")
        if asa_params["name"] == "":
            raise ValueError(f"Axiom 2: The ASA AssetName cannot be blank")

        # Axiom 3 check
        validator_balances = algo_utils.get_balances(client, ValidatorAddr)
        if txn.index not in validator_balances.keys():
            raise ValueError(
                f"Axiom 3: ValidatorAddr must be opted into the transferring ASA."
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class TavalidatorcertAlgoTransfer_Maker:
    type_name = "tavalidatorcert.algo.transfer"
    version = "000"

    def __init__(self, validator_addr: str, half_signed_cert_transfer_mtx: str):
        self.tuple = TavalidatorcertAlgoTransfer(
            ValidatorAddr=validator_addr,
            HalfSignedCertTransferMtx=half_signed_cert_transfer_mtx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: TavalidatorcertAlgoTransfer) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> TavalidatorcertAlgoTransfer:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> TavalidatorcertAlgoTransfer:
        d2 = dict(d)
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "HalfSignedCertTransferMtx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HalfSignedCertTransferMtx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return TavalidatorcertAlgoTransfer(
            ValidatorAddr=d2["ValidatorAddr"],
            HalfSignedCertTransferMtx=d2["HalfSignedCertTransferMtx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MarketPriceUnit(StrEnum):
    """
    Price unit assigned to MarketMaker MarketType

    Choices and descriptions:

      * USDPerMWh:
    """

    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit":
        """
        Returns default value USDPerMWh
        """
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]
